# futures
from __future__ import absolute_import

# standard libraries
import unittest

# third party libraries
import numpy

# local libraries
from nion.data import Image


class TestImageClass(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_create_rgba_image_from_array(self):
        image_1d_16 = numpy.zeros((16, ), dtype=numpy.double)
        image_1d_16x1 = numpy.zeros((16, 1), dtype=numpy.double)
        self.assertIsNotNone(Image.create_rgba_image_from_array(image_1d_16))
        self.assertIsNotNone(Image.create_rgba_image_from_array(image_1d_16x1))
        image_1d_rgb = numpy.zeros((16, 3), dtype=numpy.uint8)
        self.assertIsNotNone(Image.create_rgba_image_from_array(image_1d_rgb))

    def test_rebin_expand_has_even_expansion(self):
        # NOTE: statistical tests are only valid if expanded length is multiple of src length
        src = numpy.arange(0, 10)
        expanded = Image.rebin_1d(src, 50)
        self.assertAlmostEqual(numpy.mean(src), numpy.mean(expanded))
        self.assertAlmostEqual(numpy.var(src), numpy.var(expanded))
        src = numpy.arange(0, 10)
        expanded = Image.rebin_1d(src, 500)
        self.assertAlmostEqual(numpy.mean(src), numpy.mean(expanded))
        self.assertAlmostEqual(numpy.var(src), numpy.var(expanded))
        # test larger values to make sure linear mapping works (failed once)
        src = numpy.arange(0, 200)
        expanded = Image.rebin_1d(src, 600)
        self.assertAlmostEqual(numpy.mean(src), numpy.mean(expanded))
        self.assertAlmostEqual(numpy.var(src), numpy.var(expanded))

    def test_scale_cubic_is_symmetry(self):
        src1 = numpy.zeros((8, 8))
        src2 = numpy.zeros((9, 9))
        src1[3:5, 3:5] = 1
        src2[3:6, 3:6] = 1
        src1s = (Image.scaled(src1, (12, 12), 'cubic')*1000).astype(numpy.int)
        src2s = (Image.scaled(src1, (12, 12), 'cubic')*1000).astype(numpy.int)
        src1t = (Image.scaled(src1, (13, 13), 'cubic')*1000).astype(numpy.int)
        src2t = (Image.scaled(src1, (13, 13), 'cubic')*1000).astype(numpy.int)
        self.assertTrue(numpy.array_equal(src1s[0:6, 0:6], src1s[0:6, 12:5:-1]))
        self.assertTrue(numpy.array_equal(src1s[0:6, 0:6], src1s[12:5:-1, 12:5:-1]))
        self.assertTrue(numpy.array_equal(src1s[0:6, 0:6], src1s[12:5:-1, 0:6]))
        self.assertTrue(numpy.array_equal(src2s[0:6, 0:6], src2s[0:6, 12:5:-1]))
        self.assertTrue(numpy.array_equal(src2s[0:6, 0:6], src2s[12:5:-1, 12:5:-1]))
        self.assertTrue(numpy.array_equal(src2s[0:6, 0:6], src2s[12:5:-1, 0:6]))
        self.assertTrue(numpy.array_equal(src1t[0:6, 0:6], src1t[0:6, 13:6:-1]))
        self.assertTrue(numpy.array_equal(src1t[0:6, 0:6], src1t[13:6:-1, 13:6:-1]))
        self.assertTrue(numpy.array_equal(src1t[0:6, 0:6], src1t[13:6:-1, 0:6]))
        self.assertTrue(numpy.array_equal(src2t[0:6, 0:6], src2t[0:6, 13:6:-1]))
        self.assertTrue(numpy.array_equal(src2t[0:6, 0:6], src2t[13:6:-1, 13:6:-1]))
        self.assertTrue(numpy.array_equal(src2t[0:6, 0:6], src2t[13:6:-1, 0:6]))

    def test_scale_linear_is_symmetry(self):
        src1 = numpy.zeros((8, 8))
        src2 = numpy.zeros((9, 9))
        src1[3:5, 3:5] = 1
        src2[3:6, 3:6] = 1
        src1s = (Image.scaled(src1, (12, 12), 'linear')*1000).astype(numpy.int)
        src2s = (Image.scaled(src1, (12, 12), 'linear')*1000).astype(numpy.int)
        src1t = (Image.scaled(src1, (13, 13), 'linear')*1000).astype(numpy.int)
        src2t = (Image.scaled(src1, (13, 13), 'linear')*1000).astype(numpy.int)
        self.assertTrue(numpy.array_equal(src1s[0:6, 0:6], src1s[0:6, 12:5:-1]))
        self.assertTrue(numpy.array_equal(src1s[0:6, 0:6], src1s[12:5:-1, 12:5:-1]))
        self.assertTrue(numpy.array_equal(src1s[0:6, 0:6], src1s[12:5:-1, 0:6]))
        self.assertTrue(numpy.array_equal(src2s[0:6, 0:6], src2s[0:6, 12:5:-1]))
        self.assertTrue(numpy.array_equal(src2s[0:6, 0:6], src2s[12:5:-1, 12:5:-1]))
        self.assertTrue(numpy.array_equal(src2s[0:6, 0:6], src2s[12:5:-1, 0:6]))
        self.assertTrue(numpy.array_equal(src1t[0:6, 0:6], src1t[0:6, 13:6:-1]))
        self.assertTrue(numpy.array_equal(src1t[0:6, 0:6], src1t[13:6:-1, 13:6:-1]))
        self.assertTrue(numpy.array_equal(src1t[0:6, 0:6], src1t[13:6:-1, 0:6]))
        self.assertTrue(numpy.array_equal(src2t[0:6, 0:6], src2t[0:6, 13:6:-1]))
        self.assertTrue(numpy.array_equal(src2t[0:6, 0:6], src2t[13:6:-1, 13:6:-1]))
        self.assertTrue(numpy.array_equal(src2t[0:6, 0:6], src2t[13:6:-1, 0:6]))



# standard imports
import contextlib
import locale
import logging
import math
import unittest
import uuid

# third-party imports
import numpy

# local imports
from nion.data import Calibration
from nion.swift import Application
from nion.swift import DocumentController
from nion.swift import Facade
from nion.swift import Inspector
from nion.swift.model import DataItem
from nion.swift.model import DocumentModel
from nion.swift.model import Graphics
from nion.ui import TestUI
from nion.utils import Binding
from nion.utils import Observable


Facade.initialize()


class TestInspectorClass(unittest.TestCase):

    def setUp(self):
        self.app = Application.Application(TestUI.UserInterface(), set_global=False)

    def tearDown(self):
        pass

    def test_info_inspector_section_follows_title_change(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((256, 256), numpy.uint32))
            data_item.title = "Title1"
            document_model.append_data_item(data_item)
            inspector_section = Inspector.InfoInspectorSection(document_controller, data_item)
            with contextlib.closing(inspector_section):
                self.assertEqual(inspector_section.info_title_label.text, "Title1")
                data_item.title = "Title2"
                self.assertEqual(inspector_section.info_title_label.text, "Title2")

    def test_display_limits_inspector_should_bind_to_display_without_errors(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            # configure the inspectors
            document_controller.notify_focused_display_changed(data_item.primary_display_specifier.display)
            document_controller.periodic()  # force UI to update
            document_controller.notify_focused_display_changed(None)

    def test_calibration_value_and_size_float_to_string_converter_works_with_display(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            converter = Inspector.CalibratedValueFloatToStringConverter(display_specifier.data_item, display_specifier.display, 0)
            converter.convert(0.5)
            converter = Inspector.CalibratedSizeFloatToStringConverter(display_specifier.data_item, display_specifier.display, 0)
            converter.convert(0.5)

    def test_adjusting_rectangle_width_should_keep_center_constant(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            rect_graphic = Graphics.RectangleGraphic()
            rect_graphic.bounds = ((0.25, 0.25), (0.5, 0.5))
            center = rect_graphic.center
            class BoolModel(Observable.Observable):
                def __init__(self):
                    super(BoolModel, self).__init__()
                    self.dimensional_calibration_style = "relative-top-left"
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            size_width_binding = Inspector.CalibratedSizeBinding(display_specifier.data_item, 0, display_specifier.display, Binding.TuplePropertyBinding(rect_graphic, "size", 0))
            size_width_binding.update_source("0.6")
            self.assertEqual(center, rect_graphic.center)

    def test_calibration_inspector_section_binds(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            display_specifier.data_item.set_dimensional_calibration(0, Calibration.Calibration(offset=5.1, scale=1.2, units="mm"))
            inspector_section = Inspector.CalibrationsInspectorSection(document_controller, display_specifier.data_item, display_specifier.display)
            with contextlib.closing(inspector_section):
                calibration_list_widget = inspector_section._section_content_for_test.find_widget_by_id("calibration_list_widget")
                calibration_row = calibration_list_widget.find_widget_by_id("content_section").children[0]
                offset_field = calibration_row.find_widget_by_id("offset")
                scale_field = calibration_row.find_widget_by_id("scale")
                units_field = calibration_row.find_widget_by_id("units")
                self.assertEqual(offset_field.text, "5.1000")
                self.assertEqual(scale_field.text, "1.2000")
                self.assertEqual(units_field.text, u"mm")
                display_specifier.data_item.set_dimensional_calibration(0, Calibration.Calibration(offset=1.5, scale=2.1, units="mmm"))
                self.assertEqual(offset_field.text, "1.5000")
                self.assertEqual(scale_field.text, "2.1000")
                self.assertEqual(units_field.text, u"mmm")

    def test_calibration_inspector_section_follows_spatial_calibration_change(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((8, 8), numpy.uint32))
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            display_specifier.data_item.set_dimensional_calibration(0, Calibration.Calibration(units="mm"))
            display_specifier.data_item.set_dimensional_calibration(1, Calibration.Calibration(units="mm"))
            inspector_section = Inspector.CalibrationsInspectorSection(document_controller, display_specifier.data_item, display_specifier.display)
            with contextlib.closing(inspector_section):
                calibration_list_widget = inspector_section._section_content_for_test.find_widget_by_id("calibration_list_widget")
                calibration_row = calibration_list_widget.find_widget_by_id("content_section").children[0]
                units_field = calibration_row.find_widget_by_id("units")
                self.assertEqual(units_field.text, "mm")
                display_specifier.data_item.set_dimensional_calibration(0, Calibration.Calibration(units="mmm"))
                self.assertEqual(units_field.text, "mmm")

    def test_graphic_inspector_section_follows_spatial_calibration_change(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((256, 256), numpy.uint32))
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            display_specifier.display.add_graphic(Graphics.PointGraphic())
            graphic_widget = self.app.ui.create_column_widget()
            display_specifier.display.dimensional_calibration_style = "calibrated"
            display_specifier.data_item.set_dimensional_calibration(1, Calibration.Calibration(units="mm"))
            Inspector.make_point_type_inspector(document_controller, graphic_widget, display_specifier, display_specifier.display.graphics[0])
            x_widget = graphic_widget.find_widget_by_id("x")
            self.assertEqual(x_widget.text, "128.0 mm")
            display_specifier.data_item.set_dimensional_calibration(1, Calibration.Calibration(units="mmm"))
            self.assertEqual(x_widget.text, "128.0 mmm")

    def test_changing_calibration_style_to_calibrated_displays_correct_values(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((100, 100), numpy.uint32))
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            point_graphic = Graphics.PointGraphic()
            display_specifier.display.add_graphic(point_graphic)
            graphic_widget = self.app.ui.create_column_widget()
            display_specifier.display.dimensional_calibration_style = "calibrated"
            display_specifier.data_item.set_dimensional_calibration(0, Calibration.Calibration(offset=-100, scale=2, units="mm"))
            display_specifier.data_item.set_dimensional_calibration(1, Calibration.Calibration(offset=-100, scale=2, units="mm"))
            Inspector.make_point_type_inspector(document_controller, graphic_widget, display_specifier, display_specifier.display.graphics[0])
            x_widget = graphic_widget.find_widget_by_id("x")
            y_widget = graphic_widget.find_widget_by_id("y")
            self.assertEqual(x_widget.text, "0.0 mm")
            self.assertEqual(y_widget.text, "0.0 mm")
            x_widget.text = "-10"
            x_widget.on_editing_finished(x_widget.text)
            y_widget.text = "20"
            y_widget.on_editing_finished(y_widget.text)
            self.assertEqual(point_graphic.position, (0.6, 0.45))

    def test_graphic_inspector_section_displays_sensible_units(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((256, 256), numpy.uint32))
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            display_specifier.display.add_graphic(Graphics.PointGraphic())
            graphic_widget = self.app.ui.create_column_widget()
            display_specifier.display.dimensional_calibration_style = "calibrated"
            Inspector.make_point_type_inspector(document_controller, graphic_widget, display_specifier, display_specifier.display.graphics[0])
            display_specifier.data_item.set_dimensional_calibration(1, Calibration.Calibration(scale=math.sqrt(2.0), units="mm"))
            x_widget = graphic_widget.find_widget_by_id("x")
            self.assertEqual(x_widget.text, "181.0 mm")
            display_specifier.data_item.set_dimensional_calibration(1, Calibration.Calibration(scale=math.sqrt(2.0), offset=5.55555, units="mm"))
            self.assertEqual(x_widget.text, "186.6 mm")
            display_specifier.data_item.set_dimensional_calibration(1, Calibration.Calibration(scale=math.sqrt(2.0)/10, units="mm"))
            self.assertEqual(x_widget.text, "18.10 mm")
            display_specifier.data_item.set_dimensional_calibration(1, Calibration.Calibration(scale=math.sqrt(2.0)/10, offset=0.55555, units="mm"))
            self.assertEqual(x_widget.text, "18.66 mm")
            display_specifier.data_item.set_dimensional_calibration(1, Calibration.Calibration(scale=-math.sqrt(2.0)/10, units="mm"))
            self.assertEqual(x_widget.text, "-18.10 mm")

    def test_graphic_inspector_display_calibrated_length_units(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((200, 100), numpy.uint32))
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            line_region = Graphics.LineGraphic()
            line_region.start = (0, 0)
            line_region.end = (1, 1)
            display_specifier.display.add_graphic(line_region)
            graphic_widget = self.app.ui.create_column_widget()
            display_specifier.display.dimensional_calibration_style = "calibrated"
            Inspector.make_line_type_inspector(document_controller, graphic_widget, display_specifier, display_specifier.display.graphics[0])
            display_specifier.data_item.set_dimensional_calibration(0, Calibration.Calibration(units="mm"))
            display_specifier.data_item.set_dimensional_calibration(1, Calibration.Calibration(units="mm"))
            self.assertEqual(graphic_widget.find_widget_by_id("length").text, "223.607 mm")  # sqrt(100*100 + 200*200)

    def test_graphic_inspector_sets_calibrated_length_units(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((200, 100), numpy.uint32))
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            line_region = Graphics.LineGraphic()
            line_region.start = (0, 0)
            line_region.end = (0.5, 0.5)
            display_specifier.display.add_graphic(line_region)
            graphic_widget = self.app.ui.create_column_widget()
            display_specifier.display.dimensional_calibration_style = "calibrated"
            Inspector.make_line_type_inspector(document_controller, graphic_widget, display_specifier, display_specifier.display.graphics[0])
            display_specifier.data_item.set_dimensional_calibration(0, Calibration.Calibration(units="mm"))
            display_specifier.data_item.set_dimensional_calibration(1, Calibration.Calibration(units="mm"))
            length_str = "{0:g}".format(math.sqrt(100 * 100 + 200 * 200))
            length_widget = graphic_widget.find_widget_by_id("length")
            length_widget.text = length_str
            length_widget.on_editing_finished(length_str)
            self.assertAlmostEqual(line_region.end[0], 1.0, 3)
            self.assertAlmostEqual(line_region.end[1], 1.0, 3)

    def test_line_profile_inspector_display_calibrated_width_units(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((200, 100), numpy.uint32))
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            line_profile = Graphics.LineProfileGraphic()
            line_profile.start = (0, 0)
            line_profile.end = (1, 1)
            line_profile.width = 10
            display_specifier.display.add_graphic(line_profile)
            graphic_widget = self.app.ui.create_column_widget()
            display_specifier.display.dimensional_calibration_style = "calibrated"
            Inspector.make_line_profile_inspector(document_controller, graphic_widget, display_specifier, display_specifier.display.graphics[0])
            display_specifier.data_item.set_dimensional_calibration(0, Calibration.Calibration(units="mm"))
            display_specifier.data_item.set_dimensional_calibration(1, Calibration.Calibration(units="mm"))
            self.assertEqual(graphic_widget.find_widget_by_id("width").text, "10.0 mm")

    def test_float_to_string_converter_strips_units(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((256, 256), numpy.uint32))
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            display = display_specifier.display
            display.dimensional_calibration_style = "pixels-top-left"
            converter = Inspector.CalibratedValueFloatToStringConverter(data_item, display, 0)
            locale.setlocale(locale.LC_ALL, '')
            self.assertAlmostEqual(converter.convert_back("0.5"), 0.5 / 256)
            self.assertAlmostEqual(converter.convert_back(".5"), 0.5 / 256)
            self.assertAlmostEqual(converter.convert_back("00.5"), 0.5 / 256)
            self.assertAlmostEqual(converter.convert_back("0.500"), 0.5 / 256)
            self.assertAlmostEqual(converter.convert_back("0.500e0"), 0.5 / 256)
            self.assertAlmostEqual(converter.convert_back("+.5"), 0.5 / 256)
            self.assertAlmostEqual(converter.convert_back("-.5"), -0.5 / 256)
            self.assertAlmostEqual(converter.convert_back("+0.5"), 0.5 / 256)
            self.assertAlmostEqual(converter.convert_back("0.5x"), 0.5 / 256)
            self.assertAlmostEqual(converter.convert_back("x0.5"), 0.0)
            self.assertAlmostEqual(converter.convert_back(" 0.5 "), 0.5 / 256)
            self.assertAlmostEqual(converter.convert_back(""), 0.0)
            self.assertAlmostEqual(converter.convert_back("  "), 0.0)
            self.assertAlmostEqual(converter.convert_back(" x"), 0.0)
            try:
                locale.setlocale(locale.LC_ALL, 'de_DE')
                self.assertAlmostEqual(converter.convert_back("0,500"), 0.5 / 256)
                self.assertAlmostEqual(converter.convert_back("0.500"), 0.5 / 256)
            except locale.Error as e:
                pass

    def test_inspector_handles_sliced_3d_data(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((32, 32, 16)))
            document_model.append_data_item(data_item)
            display_panel.set_display_panel_data_item(data_item)
            document_controller.selected_display_panel = None
            document_controller.selected_display_panel = display_panel
            document_controller.processing_slice()
            document_model.recompute_all()
            display_panel.set_display_panel_data_item(document_model.data_items[1])
            inspector_panel = document_controller.find_dock_widget("inspector-panel").panel
            document_controller.periodic()
            self.assertTrue(len(inspector_panel._get_inspector_sections()) > 0)

    def test_slice_inspector_section_uses_correct_dimension(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((4, 4, 32), numpy.uint32))
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            display_specifier.display.slice_center = 16
            display_specifier.display.slice_width = 4
            slice_inspector_section = Inspector.SliceInspectorSection(document_controller, data_item, display_specifier.display)
            with contextlib.closing(slice_inspector_section):
                self.assertEqual(slice_inspector_section._slice_center_slider_widget.value, 16)
                self.assertEqual(slice_inspector_section._slice_center_slider_widget.maximum, 31)
                self.assertEqual(slice_inspector_section._slice_width_slider_widget.value, 4)
                self.assertEqual(slice_inspector_section._slice_width_slider_widget.maximum, 31)
                self.assertEqual(slice_inspector_section._slice_center_line_edit_widget.text, "16")
                self.assertEqual(slice_inspector_section._slice_width_line_edit_widget.text, "4")

    def test_image_display_inspector_shows_empty_fields_for_none_display_limits(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((4, 4), numpy.uint32))
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            display = display_specifier.display
            inspector_section = Inspector.ImageDisplayInspectorSection(document_controller, display)
            with contextlib.closing(inspector_section):
                display.display_limits = None
                self.assertEqual(inspector_section.display_limits_limit_low.text, None)
                self.assertEqual(inspector_section.display_limits_limit_high.text, None)
                display.display_limits = (None, None)
                self.assertEqual(inspector_section.display_limits_limit_low.text, None)
                self.assertEqual(inspector_section.display_limits_limit_high.text, None)
                display.display_limits = (1, None)
                self.assertEqual(inspector_section.display_limits_limit_low.text, "1.0000")
                self.assertEqual(inspector_section.display_limits_limit_high.text, None)
                display.display_limits = (None, 2)
                self.assertEqual(inspector_section.display_limits_limit_low.text, None)
                self.assertEqual(inspector_section.display_limits_limit_high.text, "2.0000")
                display.display_limits = (1, 2)
                self.assertEqual(inspector_section.display_limits_limit_low.text, "1.0000")
                self.assertEqual(inspector_section.display_limits_limit_high.text, "2.0000")

    def test_image_display_inspector_sets_display_limits_when_text_is_changed(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((4, 4), numpy.uint32))
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            display = display_specifier.display
            inspector_section = Inspector.ImageDisplayInspectorSection(document_controller, display)
            with contextlib.closing(inspector_section):
                self.assertEqual(display.display_limits, None)
                inspector_section.display_limits_limit_low.text = "1"
                inspector_section.display_limits_limit_low.editing_finished("1")
                self.assertEqual(display.display_limits, (1.0, None))
                inspector_section.display_limits_limit_high.text = "2"
                inspector_section.display_limits_limit_high.editing_finished("2")
                self.assertEqual(display.display_limits, (1.0, 2.0))
                inspector_section.display_limits_limit_low.text = ""
                inspector_section.display_limits_limit_low.editing_finished("")
                self.assertEqual(display.display_limits, (None, 2.0))
                inspector_section.display_limits_limit_high.text = ""
                inspector_section.display_limits_limit_high.editing_finished("")
                self.assertEqual(display.display_limits, None)

    def test_inspector_handles_deleted_data(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((32, 32)))
            document_model.append_data_item(data_item)
            document_controller.periodic()
            document_controller.select_data_item_in_data_panel(data_item=data_item)  # queues the update
            document_model.remove_data_item(data_item)  # removes item while still in queue
            document_controller.periodic()  # execute queue

    def test_inspector_handles_deleted_data_being_displayed(self):
        # if the inspector doesn't watch for the item being deleted, and the inspector's update display
        # doesn't get called during periodic before the item is deleted (which will naturally happen sometimes),
        # then there will be a pending call using a data item that doesn't exist.
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.ones((8, 8)))
            document_model.append_data_item(data_item)
            document_controller.display_data_item(DataItem.DisplaySpecifier.from_data_item(data_item))
            # document_controller.periodic()  # this makes it succeed in all cases
            document_model.remove_data_item(data_item)
            document_controller.periodic()

    def test_inspector_handles_empty_data_item(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem()
            document_model.append_data_item(data_item)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(data_item)
            document_controller.selected_display_panel = None
            document_controller.selected_display_panel = display_panel
            inspector_panel = document_controller.find_dock_widget("inspector-panel").panel
            document_controller.periodic()
            self.assertTrue(len(inspector_panel._get_inspector_sections()) > 0)

    def test_inspector_handles_all_graphics_on_1d_data(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.ones((1024, )))
            document_model.append_data_item(data_item)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(data_item)
            document_controller.selected_display_panel = None
            document_controller.selected_display_panel = display_panel
            document_controller.add_point_graphic()
            document_controller.add_line_graphic()
            document_controller.add_rectangle_graphic()
            document_controller.add_ellipse_graphic()
            document_controller.add_interval_graphic()
            data_item.displays[0].graphic_selection.clear()  # make sure all graphics show up in inspector
            self.assertTrue(len(data_item.displays[0].graphics) == 5)
            inspector_panel = document_controller.find_dock_widget("inspector-panel").panel
            document_controller.periodic()
            self.assertTrue(len(inspector_panel._get_inspector_sections()) > 0)

    def test_inspector_handles_all_graphics_on_2d_data(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.ones((256, 256)))
            document_model.append_data_item(data_item)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(data_item)
            document_controller.selected_display_panel = None
            document_controller.selected_display_panel = display_panel
            document_controller.add_point_graphic()
            document_controller.add_line_graphic()
            document_controller.add_rectangle_graphic()
            document_controller.add_ellipse_graphic()
            document_controller.add_interval_graphic()
            data_item.displays[0].graphic_selection.clear()  # make sure all graphics show up in inspector
            self.assertTrue(len(data_item.displays[0].graphics) == 5)
            inspector_panel = document_controller.find_dock_widget("inspector-panel").panel
            document_controller.periodic()
            self.assertTrue(len(inspector_panel._get_inspector_sections()) > 0)

    def test_inspector_handles_all_graphics_on_3d_data(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.ones((5, 5, 5)))
            document_model.append_data_item(data_item)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(data_item)
            document_controller.selected_display_panel = None
            document_controller.selected_display_panel = display_panel
            document_controller.add_point_graphic()
            document_controller.add_line_graphic()
            document_controller.add_rectangle_graphic()
            document_controller.add_ellipse_graphic()
            document_controller.add_interval_graphic()
            data_item.displays[0].graphic_selection.clear()  # make sure all graphics show up in inspector
            self.assertTrue(len(data_item.displays[0].graphics) == 5)
            inspector_panel = document_controller.find_dock_widget("inspector-panel").panel
            document_controller.periodic()
            self.assertTrue(len(inspector_panel._get_inspector_sections()) > 0)

    def test_updating_display_limits_on_fft_does_not_enter_update_cycle(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            d = numpy.random.randn(16, 16)
            d = (d - numpy.min(d)) / (numpy.amax(d) - numpy.amin(d)).astype(numpy.complex128)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(data_item)
            document_controller.selected_display_panel = None
            document_controller.selected_display_panel = display_panel
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            count = [0]
            def property_changed(property_name):
                if property_name == "display_limits":
                    count[0] += 1
            property_changed_listener = display_specifier.display.property_changed_event.listen(property_changed)
            document_model.recompute_all()
            document_model.recompute_all()
            self.assertEqual(count[0], 0)
            property_changed_listener.close()

    def test_rectangle_dimensions_show_calibrated_units(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.full((256, 256, 37), 0, dtype=numpy.uint32))  # z, y, x
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            display_specifier.display.add_graphic(Graphics.RectangleGraphic())
            graphic_widget = self.app.ui.create_column_widget()
            display_specifier.display.dimensional_calibration_style = "calibrated"
            display_specifier.data_item.set_dimensional_calibration(0, Calibration.Calibration(units="mm", scale=0.5))  # y
            display_specifier.data_item.set_dimensional_calibration(1, Calibration.Calibration(units="mm", scale=2.0))  # x
            Inspector.make_rectangle_type_inspector(document_controller, graphic_widget, display_specifier, display_specifier.display.graphics[0], str())
            self.assertEqual(graphic_widget.find_widget_by_id("x").text, "256.0 mm")  # x
            self.assertEqual(graphic_widget.find_widget_by_id("y").text, "64.00 mm")  # y
            self.assertEqual(graphic_widget.find_widget_by_id("width").text, "512.0 mm")  # width
            self.assertEqual(graphic_widget.find_widget_by_id("height").text, "128.00 mm")  # height

    def test_line_dimensions_show_calibrated_units(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.full((100, 100, 37), 0, dtype=numpy.uint32))  # z, y, x
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            line_graphic = Graphics.LineGraphic()
            line_graphic.start = 0.2, 0.2
            line_graphic.end = 0.8, 0.8
            display_specifier.display.add_graphic(line_graphic)
            graphic_widget = self.app.ui.create_column_widget()
            display_specifier.display.dimensional_calibration_style = "calibrated"
            display_specifier.data_item.set_dimensional_calibration(0, Calibration.Calibration(units="mm", scale=0.5))  # y
            display_specifier.data_item.set_dimensional_calibration(1, Calibration.Calibration(units="mm", scale=2.0))  # x
            Inspector.make_line_type_inspector(document_controller, graphic_widget, display_specifier, display_specifier.display.graphics[0])
            self.assertEqual(graphic_widget.find_widget_by_id("x0").text, "40.0 mm")  # x0
            self.assertEqual(graphic_widget.find_widget_by_id("y0").text, "10.00 mm")  # y0
            self.assertEqual(graphic_widget.find_widget_by_id("x1").text, "160.0 mm")  # x1
            self.assertEqual(graphic_widget.find_widget_by_id("y1").text, "40.00 mm")  # y1

    def test_point_dimensions_show_calibrated_units(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.full((256, 256, 37), 0, dtype=numpy.uint32))  # z, y, x
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            display_specifier.display.add_graphic(Graphics.PointGraphic())
            graphic_widget = self.app.ui.create_column_widget()
            display_specifier.display.dimensional_calibration_style = "calibrated"
            display_specifier.data_item.set_dimensional_calibration(0, Calibration.Calibration(units="mm", scale=0.5))  # y
            display_specifier.data_item.set_dimensional_calibration(1, Calibration.Calibration(units="mm", scale=2.0))  # x
            Inspector.make_point_type_inspector(document_controller, graphic_widget, display_specifier, display_specifier.display.graphics[0])
            self.assertEqual(graphic_widget.find_widget_by_id("x").text, "256.0 mm")  # x
            self.assertEqual(graphic_widget.find_widget_by_id("y").text, "64.00 mm")  # y

    def test_pixel_center_rounding_correct_on_odd_dimensioned_image(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.full((139, 89), 0, dtype=numpy.uint32))  # y, x
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            point_graphic = Graphics.PointGraphic()
            point_graphic.position = 0.5, 0.5
            display_specifier.display.add_graphic(point_graphic)
            graphic_widget = self.app.ui.create_column_widget()
            Inspector.make_point_type_inspector(document_controller, graphic_widget, display_specifier, display_specifier.display.graphics[0])
            display_specifier.display.dimensional_calibration_style = "pixels-center"
            self.assertEqual(graphic_widget.find_widget_by_id("x").text, "0.0")  # x
            self.assertEqual(graphic_widget.find_widget_by_id("y").text, "0.0")  # y

    def test_editing_pixel_width_on_rectangle_adjusts_rectangle_properly(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.full((100, 50), 0, dtype=numpy.uint32))  # y, x
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            point_graphic = Graphics.RectangleGraphic()
            display_specifier.display.add_graphic(point_graphic)
            graphic_widget = self.app.ui.create_column_widget()
            Inspector.make_rectangle_type_inspector(document_controller, graphic_widget, display_specifier, display_specifier.display.graphics[0], str())
            display_specifier.display.dimensional_calibration_style = "pixels-center"
            self.assertEqual(graphic_widget.find_widget_by_id("x").text, "0.0")  # x
            self.assertEqual(graphic_widget.find_widget_by_id("y").text, "0.0")  # y
            self.assertEqual(graphic_widget.find_widget_by_id("width").text, "50.0")  # width
            self.assertEqual(graphic_widget.find_widget_by_id("height").text, "100.0")  # height
            graphic_widget.find_widget_by_id("width").text = "40"
            graphic_widget.find_widget_by_id("width").editing_finished("40")
            self.assertEqual(graphic_widget.find_widget_by_id("x").text, "0.0")  # x
            self.assertEqual(graphic_widget.find_widget_by_id("y").text, "0.0")  # y
            self.assertEqual(graphic_widget.find_widget_by_id("width").text, "40.0")  # width
            self.assertEqual(graphic_widget.find_widget_by_id("height").text, "100.0")  # height

    def test_interval_dimensions_show_calibrated_units_on_single_spectrum(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.full((100, ), 0, dtype=numpy.uint32))  # time, energy
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            interval_graphic = Graphics.IntervalGraphic()
            interval_graphic.start = 0.2
            interval_graphic.end = 0.4
            display_specifier.display.add_graphic(interval_graphic)
            graphic_widget = self.app.ui.create_column_widget()
            display_specifier.display.dimensional_calibration_style = "calibrated"
            display_specifier.data_item.set_dimensional_calibration(0, Calibration.Calibration(units="eV", scale=2.0))  # energy
            Inspector.make_interval_type_inspector(document_controller, graphic_widget, display_specifier, display_specifier.display.graphics[0])
            self.assertEqual("40.0 eV", graphic_widget.find_widget_by_id("start").text)  # energy
            self.assertEqual("80.0 eV", graphic_widget.find_widget_by_id("end").text)  # energy

    def test_interval_dimensions_show_calibrated_units_on_sequence_of_spectra(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.full((3, 100), 0, dtype=numpy.uint32))  # time, energy
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            interval_graphic = Graphics.IntervalGraphic()
            interval_graphic.start = 0.2
            interval_graphic.end = 0.4
            display_specifier.display.add_graphic(interval_graphic)
            graphic_widget = self.app.ui.create_column_widget()
            display_specifier.display.dimensional_calibration_style = "calibrated"
            display_specifier.data_item.set_dimensional_calibration(0, Calibration.Calibration(units="s", scale=1.0))  # time
            display_specifier.data_item.set_dimensional_calibration(1, Calibration.Calibration(units="eV", scale=2.0))  # energy
            Inspector.make_interval_type_inspector(document_controller, graphic_widget, display_specifier, display_specifier.display.graphics[0])
            self.assertEqual("40.0 eV", graphic_widget.find_widget_by_id("start").text)  # energy
            self.assertEqual("80.0 eV", graphic_widget.find_widget_by_id("end").text)  # energy

    def test_calibration_inspector_updates_for_when_data_shape_changes(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((10, 10, 10)))
            document_model.append_data_item(data_item)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(data_item)
            inspector_panel = document_controller.find_dock_widget("inspector-panel").panel
            document_controller.periodic()
            self.assertIn(Inspector.SliceInspectorSection, (type(i) for i in inspector_panel._get_inspector_sections()))
            data_item.set_data(numpy.zeros((10, 10)))
            document_controller.periodic()
            self.assertNotIn(Inspector.SliceInspectorSection, (type(i) for i in inspector_panel._get_inspector_sections()))

    def test_calibration_inspector_updates_for_when_empty_data_item_displayed_as_line_plot_gets_data(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem()
            document_model.append_data_item(data_item)
            data_item.ensure_data_source()
            data_item.displays[0].add_graphic(Graphics.IntervalGraphic())
            data_item.displays[0].display_type = "line_plot"
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(data_item)
            document_controller.periodic()
            data_item.set_data(numpy.zeros((10, )))

    def test_inspector_updates_for_new_data_item(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            # configure workspace
            d = {"type": "splitter", "orientation": "vertical", "splits": [0.5, 0.5], "children": [
                {"type": "image", "uuid": "0569ca31-afd7-48bd-ad54-5e2bb9f21102", "identifier": "a", "selected": True},
                {"type": "image", "uuid": "acd77f9f-2f6f-4fbf-af5e-94330b73b997", "identifier": "b"}]}
            workspace_2x1 = document_controller.workspace_controller.new_workspace("2x1", d)
            document_controller.workspace_controller.change_workspace(workspace_2x1)
            # put data item in first area
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            # print("data_item {}".format(data_item))
            document_model.append_data_item(data_item)
            display_panel.set_display_panel_data_item(data_item)
            inspector_panel = document_controller.find_dock_widget("inspector-panel").panel
            document_controller.periodic()
            expected_inspector_section_count = len(inspector_panel._get_inspector_sections())
            # process to add new data item
            new_data_item = document_controller.processing_invert().data_item
            # print("new_data_item {}".format(new_data_item))
            document_model.recompute_all()
            document_controller.periodic()
            new_display_panel = document_controller.selected_display_panel
            self.assertNotEqual(new_display_panel, display_panel)
            self.assertEqual(new_display_panel.data_item, new_data_item)
            actual_inspector_section_count = len(inspector_panel._get_inspector_sections())
            self.assertEqual(expected_inspector_section_count, actual_inspector_section_count)

    def test_data_item_with_no_data_displays_as_line_plot(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem()
            document_model.append_data_item(data_item)
            data_item.ensure_data_source()
            data_item.displays[0].display_type = "line_plot"
            display_panel.set_display_panel_data_item(data_item)
            inspector_panel = document_controller.find_dock_widget("inspector-panel").panel
            document_controller.periodic()

    def test_inspector_updates_when_graphic_associated_with_pick_is_deleted(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((16, 16, 64)))
            document_model.append_data_item(data_item)
            display_panel.set_display_panel_data_item(data_item)
            inspector_panel = document_controller.find_dock_widget("inspector-panel").panel
            document_controller.periodic()
            expected_inspector_section_count = len(inspector_panel._get_inspector_sections())
            # add the point graphic, ensure that inspector is updated with just a section for point graphic
            self.assertEqual(len(document_controller.selected_display_specifier.display.graphic_selection.indexes), 0)  # make sure graphic is not selected
            document_model.get_pick_new(data_item)
            self.assertEqual(document_controller.selected_display_specifier.data_item, data_item)
            document_controller.selected_display_specifier.display.graphic_selection.add(0)
            document_controller.periodic()
            graphic_inspector_section_count = len(inspector_panel._get_inspector_sections())
            self.assertNotEqual(expected_inspector_section_count, graphic_inspector_section_count)
            # now remove the point graphic and ensure that inspector inspecting data item again
            self.assertEqual(len(document_controller.selected_display_specifier.display.graphic_selection.indexes), 1)  # make sure graphic is selected
            document_controller.remove_selected_graphics()
            document_controller.periodic()
            self.assertEqual(len(document_controller.selected_display_specifier.display.graphic_selection.indexes), 0)  # make sure graphic is not selected
            actual_inspector_section_count = len(inspector_panel._get_inspector_sections())
            self.assertEqual(expected_inspector_section_count, actual_inspector_section_count)

    def test_graphic_inspector_updates_for_when_data_shape_changes(self):
        # change from 2d item with a rectangle to a 1d item. what happens?
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((16, 16)))
            document_model.append_data_item(data_item)
            display_panel.set_display_panel_data_item(data_item)
            inspector_panel = document_controller.find_dock_widget("inspector-panel").panel
            document_controller.periodic()
            self.assertNotIn(Inspector.LinePlotDisplayInspectorSection, (type(i) for i in inspector_panel._get_inspector_sections()))
            data_item.set_data(numpy.zeros((10, )))
            document_controller.periodic()
            self.assertIn(Inspector.LinePlotDisplayInspectorSection, (type(i) for i in inspector_panel._get_inspector_sections()))
            data_item.set_data(numpy.zeros((10, 10)))
            document_controller.periodic()
            self.assertNotIn(Inspector.LinePlotDisplayInspectorSection, (type(i) for i in inspector_panel._get_inspector_sections()))

    def test_calibration_inspector_shows_correct_labels_for_1d(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((8, )))
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            inspector_section = Inspector.CalibrationsInspectorSection(document_controller, display_specifier.data_item, display_specifier.display)
            with contextlib.closing(inspector_section):
                calibration_list_widget = inspector_section._section_content_for_test.find_widget_by_id("calibration_list_widget")
                content_section = calibration_list_widget.find_widget_by_id("content_section")
                self.assertEqual(len(content_section.children), 1)
                calibration_row = content_section.children[0]
                label = calibration_row.find_widget_by_id("label")
                self.assertEqual(label.text, "Channel")

    def test_calibration_inspector_shows_correct_labels_for_2d(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((8, 8)))
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            inspector_section = Inspector.CalibrationsInspectorSection(document_controller, display_specifier.data_item, display_specifier.display)
            with contextlib.closing(inspector_section):
                calibration_list_widget = inspector_section._section_content_for_test.find_widget_by_id("calibration_list_widget")
                content_section = calibration_list_widget.find_widget_by_id("content_section")
                self.assertEqual(len(content_section.children), 2)
                self.assertEqual(content_section.children[0].find_widget_by_id("label").text, "Y")
                self.assertEqual(content_section.children[1].find_widget_by_id("label").text, "X")

    def test_calibration_inspector_shows_correct_labels_for_3d(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((8, 8, 16)))
            document_model.append_data_item(data_item)
            display_specifier = DataItem.DisplaySpecifier.from_data_item(data_item)
            inspector_section = Inspector.CalibrationsInspectorSection(document_controller, display_specifier.data_item, display_specifier.display)
            with contextlib.closing(inspector_section):
                calibration_list_widget = inspector_section._section_content_for_test.find_widget_by_id("calibration_list_widget")
                content_section = calibration_list_widget.find_widget_by_id("content_section")
                self.assertEqual(len(content_section.children), 3)
                self.assertEqual(content_section.children[0].find_widget_by_id("label").text, "0")
                self.assertEqual(content_section.children[1].find_widget_by_id("label").text, "1")
                self.assertEqual(content_section.children[2].find_widget_by_id("label").text, "2")

    def test_computation_inspector_updates_when_computation_variable_type_changes(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            computation = document_model.create_computation()
            x = computation.create_variable("x", "integral", 0)
            document_model.set_data_item_computation(data_item, computation)
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(data_item)
            inspector_panel = document_controller.find_dock_widget("inspector-panel").panel
            document_controller.periodic()
            inspector_section = next(x for x in inspector_panel._get_inspector_sections() if isinstance(x, Inspector.ComputationInspectorSection))
            line_edit_widget1 = inspector_section._variables_column_widget.children[0].content_widget.find_widget_by_id("value")
            line_edit_widget1.editing_finished("1")
            self.assertEqual(x.value, 1)
            x.value_type = "real"
            document_controller.periodic()
            line_edit_widget2 = inspector_section._variables_column_widget.children[0].content_widget.find_widget_by_id("value")
            line_edit_widget2.editing_finished("1.1")
            self.assertEqual(x.value, 1.1)

    def test_computation_inspector_handles_computation_variable_checkbox_and_undo_redo_cycle(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            new_data_item = document_model.get_transpose_flip_new(data_item)
            document_model.recompute_all()
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(new_data_item)
            inspector_panel = document_controller.find_dock_widget("inspector-panel").panel
            document_controller.periodic()
            computation = document_model.get_data_item_computation(new_data_item)
            self.assertFalse(computation._get_variable("do_transpose").value)
            inspector_section = next(x for x in inspector_panel._get_inspector_sections() if isinstance(x, Inspector.ComputationInspectorSection))
            cb_widget = inspector_section.find_widget_by_id("value")
            cb_widget.on_checked_changed(True)
            self.assertTrue(computation._get_variable("do_transpose").value)
            document_controller.handle_undo()
            self.assertFalse(computation._get_variable("do_transpose").value)
            document_controller.handle_redo()
            self.assertTrue(computation._get_variable("do_transpose").value)

    def test_computation_inspector_handles_computation_variable_slider_and_undo_redo_cycle(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            new_data_item = document_model.get_gaussian_blur_new(data_item)
            document_model.recompute_all()
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(new_data_item)
            inspector_panel = document_controller.find_dock_widget("inspector-panel").panel
            document_controller.periodic()
            computation = document_model.get_data_item_computation(new_data_item)
            old_sigma = computation._get_variable("sigma").value
            inspector_section = next(x for x in inspector_panel._get_inspector_sections() if isinstance(x, Inspector.ComputationInspectorSection))
            slider_widget = inspector_section.find_widget_by_id("value")
            slider_widget.on_value_changed(0)
            self.assertEqual(0, computation._get_variable("sigma").value)
            document_controller.handle_undo()
            self.assertEqual(old_sigma, computation._get_variable("sigma").value)
            document_controller.handle_redo()
            self.assertEqual(0, computation._get_variable("sigma").value)

    def test_computation_inspector_handles_computation_variable_int_and_undo_redo_cycle(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            new_data_item = document_model.get_histogram_new(data_item)
            document_model.recompute_all()
            display_panel = document_controller.selected_display_panel
            display_panel.set_display_panel_data_item(new_data_item)
            inspector_panel = document_controller.find_dock_widget("inspector-panel").panel
            document_controller.periodic()
            computation = document_model.get_data_item_computation(new_data_item)
            old_bins = computation._get_variable("bins").value
            inspector_section = next(x for x in inspector_panel._get_inspector_sections() if isinstance(x, Inspector.ComputationInspectorSection))
            field_widget = inspector_section.find_widget_by_id("value")
            field_widget.on_editing_finished("100")
            self.assertEqual(100, computation._get_variable("bins").value)
            document_controller.handle_undo()
            self.assertEqual(old_bins, computation._get_variable("bins").value)
            document_controller.handle_redo()
            self.assertEqual(100, computation._get_variable("bins").value)

    def test_computation_inspector_handles_computation_variable_specifier_and_undo_redo_cycle(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((10, 10)))
            document_model.append_data_item(data_item)
            new_data_item = document_model.get_invert_new(data_item)
            document_model.recompute_all()
            computation = document_model.get_data_item_computation(new_data_item)
            self.assertEqual(data_item, list(computation._get_variable("src").bound_item.base_objects)[0])
            specifier = {"type": "data_item", "version": 1, "uuid": str(uuid.uuid4())}
            command = Inspector.ChangeComputationVariableCommand(document_model, computation, computation._get_variable("src"), specifier=specifier)
            command.perform()
            document_controller.push_undo_command(command)
            self.assertIsNone(computation._get_variable("src").bound_item)
            self.assertTrue(document_controller._undo_stack.can_undo)
            document_controller.handle_undo()
            self.assertEqual(data_item, list(computation._get_variable("src").bound_item.base_objects)[0])
            document_controller.handle_redo()
            self.assertIsNone(computation._get_variable("src").bound_item)

    def test_change_property_command_multiple_undo(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((256, 256), numpy.uint32))
            document_model.append_data_item(data_item)
            session_metadata = data_item.session_metadata
            session_metadata["site"] = "S"
            session_metadata["instrument"] = "I"
            session_metadata["task"] = "T"
            data_item.session_metadata = session_metadata
            session_metadata = data_item.session_metadata
            session_metadata["site"] = "SS"
            command = Inspector.ChangePropertyCommand(document_model, data_item, "session_metadata", session_metadata)
            command.perform()
            document_controller.push_undo_command(command)
            session_metadata = data_item.session_metadata
            session_metadata["instrument"] = "II"
            command = Inspector.ChangePropertyCommand(document_model, data_item, "session_metadata", session_metadata)
            command.perform()
            document_controller.push_undo_command(command)
            session_metadata = data_item.session_metadata
            session_metadata["task"] = "TT"
            command = Inspector.ChangePropertyCommand(document_model, data_item, "session_metadata", session_metadata)
            command.perform()
            document_controller.push_undo_command(command)
            self.assertEqual(1, document_controller._undo_stack._undo_count)
            self.assertEqual({"site": "SS", "instrument": "II", "task": "TT"}, data_item.session_metadata)
            document_controller.handle_undo()
            self.assertEqual(0, document_controller._undo_stack._undo_count)
            self.assertEqual({"site": "S", "instrument": "I", "task": "T"}, data_item.session_metadata)
            document_controller.handle_redo()
            self.assertEqual(1, document_controller._undo_stack._undo_count)
            self.assertEqual({"site": "SS", "instrument": "II", "task": "TT"}, data_item.session_metadata)

    def test_change_property_command_with_different_properties_multiple_undo(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((256, 256), numpy.uint32))
            document_model.append_data_item(data_item)
            command = Inspector.ChangePropertyCommand(document_model, data_item, "title", "T")
            command.perform()
            document_controller.push_undo_command(command)
            command = Inspector.ChangePropertyCommand(document_model, data_item, "caption", "C")
            command.perform()
            document_controller.push_undo_command(command)
            self.assertEqual(2, document_controller._undo_stack._undo_count)
            document_controller.handle_undo()
            self.assertEqual("T", data_item.title)
            document_controller.handle_undo()
            self.assertEqual(0, document_controller._undo_stack._undo_count)

    def test_change_intensity_calibration_command(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((256, 256), numpy.uint32))
            document_model.append_data_item(data_item)
            command = Inspector.ChangeIntensityCalibrationCommand(document_model, data_item, Calibration.Calibration(scale=3))
            command.perform()
            document_controller.push_undo_command(command)
            self.assertEqual(Calibration.Calibration(scale=3), data_item.intensity_calibration)
            document_controller.handle_undo()
            self.assertEqual(Calibration.Calibration(), data_item.intensity_calibration)
            document_controller.handle_redo()
            self.assertEqual(Calibration.Calibration(scale=3), data_item.intensity_calibration)

    def test_change_dimensional_calibrations_command(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((256, 256), numpy.uint32))
            document_model.append_data_item(data_item)
            command = Inspector.ChangeDimensionalCalibrationsCommand(document_model, data_item, [Calibration.Calibration(scale=2), Calibration.Calibration(scale=3)])
            command.perform()
            document_controller.push_undo_command(command)
            self.assertEqual([Calibration.Calibration(scale=2), Calibration.Calibration(scale=3)], data_item.dimensional_calibrations)
            document_controller.handle_undo()
            self.assertEqual([Calibration.Calibration(), Calibration.Calibration()], data_item.dimensional_calibrations)
            document_controller.handle_redo()
            self.assertEqual([Calibration.Calibration(scale=2), Calibration.Calibration(scale=3)], data_item.dimensional_calibrations)


if __name__ == '__main__':
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



# futures
from __future__ import absolute_import

# standard libraries
import gettext
import weakref

# third party libraries
import numpy

# local libraries
from nion.swift import Decorators
from nion.swift import Panel
from nion.ui import CanvasItem
from nion.utils import Geometry

_ = gettext.gettext


class ToolbarPanel(Panel.Panel):

    def __init__(self, document_controller, panel_id, properties):
        super(ToolbarPanel, self).__init__(document_controller, panel_id, _("Toolbar"))

        self.widget = self.ui.create_column_widget()

        toolbar_row_widget = self.ui.create_row_widget()

        # see https://www.iconfinder.com

        ui = document_controller.ui

        document_controller_weak_ref = weakref.ref(document_controller)

        icon_size = Geometry.IntSize(height=24, width=32)
        border_color = "#CCC"

        margins = Geometry.Margins(left=2, right=2, top=3, bottom=3)

        tool_palette_grid_canvas_item = CanvasItem.CanvasItemComposition()
        tool_palette_grid_canvas_item.layout = CanvasItem.CanvasItemGridLayout(size=Geometry.IntSize(height=2, width=6), margins=margins)

        pointer_tool_button = CanvasItem.BitmapButtonCanvasItem(ui.load_rgba_data_from_file(Decorators.relative_file(__file__, "resources/pointer_icon.png")), border_color=border_color)
        pointer_tool_button.size = icon_size
        pointer_tool_button.tool_tip = _("Pointer tool for selecting graphics")

        hand_tool_button = CanvasItem.BitmapButtonCanvasItem(ui.load_rgba_data_from_file(Decorators.relative_file(__file__, "resources/hand_icon.png")), border_color=border_color)
        hand_tool_button.size = icon_size
        hand_tool_button.tool_tip = _("Hand tool for dragging images within panel")

        line_tool_button = CanvasItem.BitmapButtonCanvasItem(ui.load_rgba_data_from_file(Decorators.relative_file(__file__, "resources/line_icon.png")), border_color=border_color)
        line_tool_button.size = icon_size
        line_tool_button.tool_tip = _("Line tool for making line regions on images")

        rectangle_tool_button = CanvasItem.BitmapButtonCanvasItem(ui.load_rgba_data_from_file(Decorators.relative_file(__file__, "resources/rectangle_icon.png")), border_color=border_color)
        rectangle_tool_button.size = icon_size
        rectangle_tool_button.tool_tip = _("Rectangle tool for making rectangle regions on images")

        ellipse_tool_button = CanvasItem.BitmapButtonCanvasItem(ui.load_rgba_data_from_file(Decorators.relative_file(__file__, "resources/ellipse_icon.png")), border_color=border_color)
        ellipse_tool_button.size = icon_size
        ellipse_tool_button.tool_tip = _("Ellipse tool for making ellipse regions on images")

        point_tool_button = CanvasItem.BitmapButtonCanvasItem(ui.load_rgba_data_from_file(Decorators.relative_file(__file__, "resources/point_icon.png")), border_color=border_color)
        point_tool_button.size = icon_size
        point_tool_button.tool_tip = _("Point tool for making point regions on images")

        line_profile_tool_button = CanvasItem.BitmapButtonCanvasItem(ui.load_rgba_data_from_file(Decorators.relative_file(__file__, "resources/line_profile_icon.png")), border_color=border_color)
        line_profile_tool_button.size = icon_size
        line_profile_tool_button.tool_tip = _("Line profile tool for making line profiles on images")

        interval_tool_button = CanvasItem.BitmapButtonCanvasItem(ui.load_rgba_data_from_file(Decorators.relative_file(__file__, "resources/interval_icon.png")), border_color=border_color)
        interval_tool_button.size = icon_size
        interval_tool_button.tool_tip = _("Interval tool for making intervals on line plots")

        spot_tool_button = CanvasItem.BitmapButtonCanvasItem(ui.load_rgba_data_from_file(Decorators.relative_file(__file__, "resources/spot_icon.png")), border_color=border_color)
        spot_tool_button.size = icon_size
        spot_tool_button.tool_tip = _("Spot tool for making spots on radial images")

        wedge_tool_button = CanvasItem.BitmapButtonCanvasItem(ui.load_rgba_data_from_file(Decorators.relative_file(__file__, "resources/wedge_icon.png")), border_color=border_color)
        wedge_tool_button.size = icon_size
        wedge_tool_button.tool_tip = _("Angle tool for specifying arcs on radial images")

        ring_tool_button = CanvasItem.BitmapButtonCanvasItem( ui.load_rgba_data_from_file(Decorators.relative_file(__file__, "resources/annular_ring.png")), border_color=border_color)
        ring_tool_button.size = icon_size
        ring_tool_button.tool_tip = _("Ring tool for selecting ranges on radial images")

        tool_palette_grid_canvas_item.add_canvas_item(pointer_tool_button, Geometry.IntPoint(x=0, y=0))
        tool_palette_grid_canvas_item.add_canvas_item(hand_tool_button, Geometry.IntPoint(x=0, y=1))
        tool_palette_grid_canvas_item.add_canvas_item(line_tool_button, Geometry.IntPoint(x=1, y=0))
        tool_palette_grid_canvas_item.add_canvas_item(ellipse_tool_button, Geometry.IntPoint(x=1, y=1))
        tool_palette_grid_canvas_item.add_canvas_item(rectangle_tool_button, Geometry.IntPoint(x=2, y=0))
        tool_palette_grid_canvas_item.add_canvas_item(point_tool_button, Geometry.IntPoint(x=2, y=1))
        tool_palette_grid_canvas_item.add_canvas_item(line_profile_tool_button, Geometry.IntPoint(x=3, y=0))
        tool_palette_grid_canvas_item.add_canvas_item(interval_tool_button, Geometry.IntPoint(x=3, y=1))
        tool_palette_grid_canvas_item.add_canvas_item(spot_tool_button, Geometry.IntPoint(x=4, y=0))
        tool_palette_grid_canvas_item.add_canvas_item(wedge_tool_button, Geometry.IntPoint(x=4, y=1))
        tool_palette_grid_canvas_item.add_canvas_item(ring_tool_button, Geometry.IntPoint(x=5, y=0))

        modes = "pointer", "hand", "line", "rectangle", "ellipse", "point", "line-profile", "interval", "spot", "wedge", "ring"
        self.__tool_button_group = CanvasItem.RadioButtonGroup([pointer_tool_button, hand_tool_button, line_tool_button, rectangle_tool_button, ellipse_tool_button, point_tool_button, line_profile_tool_button, interval_tool_button, spot_tool_button, wedge_tool_button, ring_tool_button])
        def tool_mode_changed(tool_mode):
            self.__tool_button_group.current_index = modes.index(tool_mode)
        self.__tool_mode_changed_event_listener = document_controller.tool_mode_changed_event.listen(tool_mode_changed)
        self.__tool_button_group.current_index = modes.index(document_controller.tool_mode)
        self.__tool_button_group.on_current_index_changed = lambda index: setattr(document_controller_weak_ref(), "tool_mode", modes[index])
        tool_mode_changed(document_controller.tool_mode)

        new_group_button = self.ui.create_push_button_widget()
        new_group_button.tool_tip = _("New Group")
        new_group_button.icon = self.ui.load_rgba_data_from_file(Decorators.relative_file(__file__, "resources/new_group_icon.png"))
        new_group_button.on_clicked = lambda: document_controller_weak_ref()._add_group_action.trigger()

        delete_button = self.ui.create_push_button_widget()
        delete_button.tool_tip = _("Delete")
        delete_button.icon = self.ui.load_rgba_data_from_file(Decorators.relative_file(__file__, "resources/delete_icon.png"))
        delete_button.on_clicked = lambda: document_controller_weak_ref()._delete_action.trigger()

        export_button = self.ui.create_push_button_widget()
        export_button.tool_tip = _("Export")
        export_button.icon = self.ui.load_rgba_data_from_file(Decorators.relative_file(__file__, "resources/export_icon.png"))
        export_button.on_clicked = lambda: document_controller_weak_ref()._export_action.trigger()

        view_palette_grid_canvas_item = CanvasItem.CanvasItemComposition()
        view_palette_grid_canvas_item.layout = CanvasItem.CanvasItemGridLayout(size=Geometry.IntSize(height=2, width=2), margins=margins)

        fit_view_button = CanvasItem.BitmapButtonCanvasItem(ui.load_rgba_data_from_file(Decorators.relative_file(__file__, "resources/fit_icon.png")), border_color=border_color)
        fit_view_button.size = icon_size
        fit_view_button.on_button_clicked = lambda: document_controller_weak_ref()._fit_view_action.trigger()
        fit_view_button.tool_tip = _("Zoom to fit to enclosing space")

        fill_view_button = CanvasItem.BitmapButtonCanvasItem(ui.load_rgba_data_from_file(Decorators.relative_file(__file__, "resources/fill_icon.png")), border_color=border_color)
        fill_view_button.size = icon_size
        fill_view_button.on_button_clicked = lambda: document_controller_weak_ref()._fill_view_action.trigger()
        fill_view_button.tool_tip = _("Zoom to fill enclosing space")

        one_to_one_view_button = CanvasItem.BitmapButtonCanvasItem(ui.load_rgba_data_from_file(Decorators.relative_file(__file__, "resources/1x1_icon.png")), border_color=border_color)
        one_to_one_view_button.size = icon_size
        one_to_one_view_button.on_button_clicked = lambda: document_controller_weak_ref()._one_to_one_view_action.trigger()
        one_to_one_view_button.tool_tip = _("Zoom to one image pixel per screen pixel")

        two_to_one_view_button = CanvasItem.BitmapButtonCanvasItem(ui.load_rgba_data_from_file(Decorators.relative_file(__file__, "resources/2x1_icon.png")), border_color=border_color)
        two_to_one_view_button.size = icon_size
        two_to_one_view_button.on_button_clicked = lambda: document_controller_weak_ref()._two_to_one_view_action.trigger()
        two_to_one_view_button.tool_tip = _("Zoom to two image pixels per screen pixel")

        view_palette_grid_canvas_item.add_canvas_item(fit_view_button, Geometry.IntPoint(x=0, y=0))
        view_palette_grid_canvas_item.add_canvas_item(fill_view_button, Geometry.IntPoint(x=0, y=1))
        view_palette_grid_canvas_item.add_canvas_item(one_to_one_view_button, Geometry.IntPoint(x=1, y=0))
        view_palette_grid_canvas_item.add_canvas_item(two_to_one_view_button, Geometry.IntPoint(x=1, y=1))

        toggle_filter_button = self.ui.create_push_button_widget()
        toggle_filter_button.tool_tip = _("Toggle Filter Panel")
        toggle_filter_button.icon = self.ui.load_rgba_data_from_file(Decorators.relative_file(__file__, "resources/filter_icon.png"))
        toggle_filter_button.on_clicked = lambda: document_controller_weak_ref()._toggle_filter_action.trigger()

        tool_palette_widget = ui.create_canvas_widget(properties={"height": 54, "width": 164})
        tool_palette_widget.canvas_item.add_canvas_item(tool_palette_grid_canvas_item)

        tool_group_widget = self.ui.create_row_widget()
        tool_group_widget.add(tool_palette_widget)

        commands_group_widget = self.ui.create_row_widget()
        commands_group_widget.add(new_group_button)
        commands_group_widget.add(delete_button)
        commands_group_widget.add(export_button)

        view_palette_widget = ui.create_canvas_widget(properties={"height": 54, "width": 68})
        view_palette_widget.canvas_item.add_canvas_item(view_palette_grid_canvas_item)

        view_group_widget = self.ui.create_row_widget()
        view_group_widget.add(view_palette_widget)

        filter_group_widget = self.ui.create_row_widget()
        filter_group_widget.add(toggle_filter_button)

        toolbar_row_widget.add_spacing(12)
        toolbar_row_widget.add(tool_group_widget)
        toolbar_row_widget.add_spacing(12)
        toolbar_row_widget.add(commands_group_widget)
        toolbar_row_widget.add_spacing(12)
        toolbar_row_widget.add(view_group_widget)
        toolbar_row_widget.add_spacing(12)
        toolbar_row_widget.add(filter_group_widget)
        toolbar_row_widget.add_spacing(12)
        toolbar_row_widget.add_stretch()

        self.widget.add(toolbar_row_widget)

    def close(self):
        self.__tool_mode_changed_event_listener.close()
        self.__tool_mode_changed_event_listener = None



# standard libraries
import collections
import copy
import math
import threading

# third party libraries
import numpy

# local libraries
from nion.data import Calibration
from nion.data import DataAndMetadata
from nion.data import Image
from nion.swift import LineGraphCanvasItem
from nion.swift import Undo
from nion.swift.model import Graphics
from nion.swift.model import Utility
from nion.ui import CanvasItem
from nion.utils import Geometry


class LinePlotCanvasItemMapping:

    def __init__(self, scale, plot_rect, left_channel, right_channel):
        self.__scale = scale
        self.__plot_rect = plot_rect
        self.__left_channel = left_channel
        self.__right_channel = right_channel
        self.__drawn_channel_per_pixel = float(right_channel - left_channel) / plot_rect.width

    def map_point_widget_to_channel_norm(self, pos):
        return (self.__left_channel + (pos.x - self.__plot_rect.left) * self.__drawn_channel_per_pixel) / self.__scale

    def map_point_channel_norm_to_widget(self, x):
        return (x * self.__scale - self.__left_channel) / self.__drawn_channel_per_pixel + self.__plot_rect.left

    def map_point_channel_norm_to_channel(self, x):
        return x * self.__scale


class LinePlotCanvasItemDelegate:
    # interface must be implemented by the delegate

    def begin_mouse_tracking(self) -> None: ...

    def end_mouse_tracking(self, undo_command) -> None: ...

    def delete_key_pressed(self) -> None: ...

    def enter_key_pressed(self) -> None: ...

    def cursor_changed(self, pos): ...

    def update_display_properties(self, display_properties: dict) -> None: ...

    def create_change_display_command(self, *, command_id: str=None, is_mergeable: bool=False) -> Undo.UndoableCommand: ...

    def create_change_graphics_command(self) -> Undo.UndoableCommand: ...

    def push_undo_command(self, command: Undo.UndoableCommand) -> None: ...

    def add_index_to_selection(self, index: int) -> None: ...

    def remove_index_from_selection(self, index: int) -> None: ...

    def set_selection(self, index: int) -> None: ...

    def clear_selection(self) -> None: ...

    def add_and_select_region(self, region: Graphics.Graphic) -> Undo.UndoableCommand: ...

    def nudge_selected_graphics(self, mapping, delta) -> None: ...

    def update_graphics(self, widget_mapping, graphic_drag_items, graphic_drag_part, graphic_part_data, graphic_drag_start_pos, pos, modifiers) -> None: ...

    def show_display_context_menu(self, gx, gy) -> bool: ...

    @property
    def tool_mode(self) -> str: return str()

    @tool_mode.setter
    def tool_mode(self, value: str) -> None: ...


class LinePlotCanvasItem(CanvasItem.LayerCanvasItem):
    """Display a line plot.

    The layout is dependent on the axes due to the dependence on the width of the text labels in the vertical axis. The
    layout is handled by the canvas items after `refresh_layout` is called, which is only called when a change in axes
    will result in different text widths.

    The graph for each data layer is dependent on the layout, data and axes. It is possible for the axes to change
    without the layout changing if, for example, only the calibrations change but the axes stay the same. If the axes or
    the data changes, however, the plot is always redrawn. Drawing is handled by the canvas item after `update` is
    called.

    axis_y = AxisYFunction(data)
    layout = LayoutFunction(axis_y)
    axis_x = AxesXFunction(layout, data)
    plot = PlotFunction(layout, data, axis_y, axis_x)
    painting = PaintFunction(layout, plot)
    """

    def __init__(self, get_font_metrics_fn, delegate: LinePlotCanvasItemDelegate, event_loop, draw_background: bool=True):
        super().__init__()

        self.__get_font_metrics_fn = get_font_metrics_fn
        self.delegate = delegate

        self.wants_mouse_events = True

        font_size = 12

        self.__closing_lock = threading.RLock()
        self.__closed = False

        self.__line_graph_xdata_list = list()

        self.__is_data1 = False
        self.__xdata_list = list()
        self.__last_data_list = list()

        # frame rate
        self.__display_frame_rate_id = None
        self.__display_frame_rate_last_index = 0

        # child displays
        self.__display_listeners = list()

        self.__line_graph_area_stack = CanvasItem.CanvasItemComposition()
        self.__line_graph_background_canvas_item = LineGraphCanvasItem.LineGraphBackgroundCanvasItem()
        self.__line_graph_stack = CanvasItem.CanvasItemComposition()
        self.__line_graph_regions_canvas_item = LineGraphCanvasItem.LineGraphRegionsCanvasItem()
        self.__line_graph_legend_canvas_item = LineGraphCanvasItem.LineGraphLegendCanvasItem(get_font_metrics_fn)
        self.__line_graph_frame_canvas_item = LineGraphCanvasItem.LineGraphFrameCanvasItem()
        self.__line_graph_area_stack.add_canvas_item(self.__line_graph_background_canvas_item)
        self.__line_graph_area_stack.add_canvas_item(self.__line_graph_stack)
        self.__line_graph_area_stack.add_canvas_item(self.__line_graph_regions_canvas_item)
        self.__line_graph_area_stack.add_canvas_item(self.__line_graph_legend_canvas_item)
        self.__line_graph_area_stack.add_canvas_item(self.__line_graph_frame_canvas_item)

        for i in range(16):
            self.__line_graph_stack.add_canvas_item(LineGraphCanvasItem.LineGraphCanvasItem())

        self.__line_graph_vertical_axis_label_canvas_item = LineGraphCanvasItem.LineGraphVerticalAxisLabelCanvasItem()
        self.__line_graph_vertical_axis_scale_canvas_item = LineGraphCanvasItem.LineGraphVerticalAxisScaleCanvasItem()
        self.__line_graph_vertical_axis_ticks_canvas_item = LineGraphCanvasItem.LineGraphVerticalAxisTicksCanvasItem()
        self.__line_graph_vertical_axis_group_canvas_item = CanvasItem.CanvasItemComposition()
        self.__line_graph_vertical_axis_group_canvas_item.layout = CanvasItem.CanvasItemRowLayout(spacing=4)
        self.__line_graph_vertical_axis_group_canvas_item.add_canvas_item(self.__line_graph_vertical_axis_label_canvas_item)
        self.__line_graph_vertical_axis_group_canvas_item.add_canvas_item(self.__line_graph_vertical_axis_scale_canvas_item)
        self.__line_graph_vertical_axis_group_canvas_item.add_canvas_item(self.__line_graph_vertical_axis_ticks_canvas_item)

        self.__line_graph_horizontal_axis_label_canvas_item = LineGraphCanvasItem.LineGraphHorizontalAxisLabelCanvasItem()
        self.__line_graph_horizontal_axis_scale_canvas_item = LineGraphCanvasItem.LineGraphHorizontalAxisScaleCanvasItem()
        self.__line_graph_horizontal_axis_ticks_canvas_item = LineGraphCanvasItem.LineGraphHorizontalAxisTicksCanvasItem()
        self.line_graph_horizontal_axis_group_canvas_item = CanvasItem.CanvasItemComposition()
        self.line_graph_horizontal_axis_group_canvas_item.layout = CanvasItem.CanvasItemColumnLayout(spacing=4)
        self.line_graph_horizontal_axis_group_canvas_item.add_canvas_item(self.__line_graph_horizontal_axis_ticks_canvas_item)
        self.line_graph_horizontal_axis_group_canvas_item.add_canvas_item(self.__line_graph_horizontal_axis_scale_canvas_item)
        self.line_graph_horizontal_axis_group_canvas_item.add_canvas_item(self.__line_graph_horizontal_axis_label_canvas_item)

        # create the grid item holding the line graph and each axes label
        line_graph_group_canvas_item = CanvasItem.CanvasItemComposition()
        margins = Geometry.Margins(left=6, right=12, top=font_size + 4, bottom=6)
        line_graph_group_canvas_item.layout = CanvasItem.CanvasItemGridLayout(Geometry.IntSize(2, 2), margins=margins)
        line_graph_group_canvas_item.add_canvas_item(self.__line_graph_vertical_axis_group_canvas_item, Geometry.IntPoint(x=0, y=0))
        line_graph_group_canvas_item.add_canvas_item(self.__line_graph_area_stack, Geometry.IntPoint(x=1, y=0))
        line_graph_group_canvas_item.add_canvas_item(self.line_graph_horizontal_axis_group_canvas_item, Geometry.IntPoint(x=1, y=1))

        # draw the background
        line_graph_background_canvas_item = CanvasItem.CanvasItemComposition()
        #line_graph_background_canvas_item.sizing.minimum_aspect_ratio = 1.5  # note: no maximum aspect ratio; line plot looks nice wider.
        line_graph_background_canvas_item.add_canvas_item(CanvasItem.BackgroundCanvasItem("#FFF"))
        line_graph_background_canvas_item.add_canvas_item(line_graph_group_canvas_item)

        # canvas items get added back to front
        # create the child canvas items
        # the background
        self.add_canvas_item(CanvasItem.BackgroundCanvasItem())
        self.add_canvas_item(line_graph_background_canvas_item)

        # used for tracking undo
        self.__undo_command = None

        # used for dragging graphic items
        self.__graphic_drag_items = []
        self.__graphic_drag_item = None
        self.__graphic_part_data = {}
        self.__graphic_drag_indexes = []
        self.__last_mouse = None
        self.__mouse_in = False
        self.__tracking_selections = False
        self.__tracking_horizontal = False
        self.__tracking_vertical = False

        self.__axes = None

        self.__data_scale = None
        self.__dimensional_calibration = None
        self.__displayed_dimensional_calibration = None
        self.__intensity_calibration = None
        self.__dimensional_calibration_style = None
        self.__y_min = None
        self.__y_max = None
        self.__y_style = None
        self.__left_channel = None
        self.__right_channel = None
        self.__legend_labels = None

        self.__graphics = list()
        self.__graphic_selection = None

    def close(self):
        # call super
        with self.__closing_lock:
            self.__closed = True
        super().close()

    @property
    def line_graph_stack(self):
        return self.__line_graph_stack

    @property
    def line_graph_canvas_item(self):
        return self.__line_graph_stack.canvas_items[0] if (self.__line_graph_stack and len(self.__line_graph_xdata_list) > 0) else None

    # for testing
    @property
    def _axes(self) -> LineGraphCanvasItem.LineGraphAxes:
        return self.__axes

    @property
    def default_aspect_ratio(self):
        return (1 + 5 ** 0.5) / 2  # golden ratio

    def display_inserted(self, display, index):

        def display_changed():
            # update the data
            display_values = display.get_calculated_display_values()
            display_data_and_metadata = display_values.display_data_and_metadata
            if display_data_and_metadata:
                self.__xdata_list[index] = display_data_and_metadata
            else:
                self.__xdata_list[index] = None
            # update the cursor info
            self.__update_cursor_info()
            # mark for update. prepare display will mark children for update if necesssary.
            self.update()

        self.__xdata_list.insert(index, None)
        self.__display_listeners.insert(index, display.display_changed_event.listen(display_changed))

        display_changed()

    def display_removed(self, display, index):
        del self.__xdata_list[index]
        self.__display_listeners[index].close()
        del self.__display_listeners[index]
        # update the cursor info
        self.delegate.cursor_changed(None)
        # mark for update. prepare display will mark children for update if necesssary.
        self.update()

    def display_rgba_changed(self, display, display_values):
        # when the display rgba data changes, no need to do anything
        pass

    def display_data_and_metadata_changed(self, display, display_values):
        # when the data changes, update the display.
        self.update_display_values(display, display_values)

    def update_display_values(self, display, display_values):
        """Update the display values. Called from display panel.

        This method saves the display values and data and triggers an update. It should be as fast as possible.

        As a layer, this canvas item will respond to the update by calling prepare_render on the layer's rendering
        thread. Prepare render will call prepare_display which will construct new axes and update all of the constituent
        canvas items such as the axes labels and the graph layers. Each will trigger its own update if its inputs have
        changed.

        The inefficiencies in this process are that the layer must re-render on each call to this function. There is
        also a cost within the constituent canvas items to check whether the axes or their data has changed.

        When the display is associated with a single data item, the data will be
        """

        # may be called from thread; prevent a race condition with closing.
        with self.__closing_lock:
            if self.__closed:
                return

            dimensional_calibrations = display.dimensional_calibrations
            displayed_dimensional_scales = display.displayed_dimensional_scales
            displayed_dimensional_calibrations = display.displayed_dimensional_calibrations
            self.__data_scale = displayed_dimensional_scales[-1] if len(displayed_dimensional_scales) > 0 else 1
            self.__dimensional_calibration = dimensional_calibrations[-1] if dimensional_calibrations else None
            self.__displayed_dimensional_calibration = displayed_dimensional_calibrations[-1] if len(displayed_dimensional_calibrations) > 0 else Calibration.Calibration(scale=displayed_dimensional_scales[-1])
            self.__intensity_calibration = display.displayed_intensity_calibration
            self.__dimensional_calibration_style = display.get_calibration_style_for_id(display.dimensional_calibration_style)
            self.__y_min = display.y_min
            self.__y_max = display.y_max
            self.__y_style = display.y_style
            self.__left_channel = display.left_channel
            self.__right_channel = display.right_channel
            self.__legend_labels = display.legend_labels

            display_data_and_metadata = display_values.display_data_and_metadata
            if display_data_and_metadata:
                # store the pending display parameters
                self.__is_data1 = True
                self.__xdata_list = [display_data_and_metadata]
                self.__update_frame(display.data_and_metadata_for_display_panel.metadata)
            else:
                if self.__is_data1:
                    self.__is_data1 = False
                    self.__xdata_list = list()

            # update the cursor info
            self.__update_cursor_info()

            # mark for update. prepare display will mark children for update if necesssary.
            self.update()

    def __update_frame(self, metadata):
        # update frame rate info
        if self.__display_frame_rate_id:
            frame_index = metadata.get("hardware_source", dict()).get("frame_index", 0)
            if frame_index != self.__display_frame_rate_last_index:
                Utility.fps_tick("frame_"+self.__display_frame_rate_id)
                self.__display_frame_rate_last_index = frame_index

    def update(self):
        if self.__display_frame_rate_id:
            if len(self.__xdata_list) != len(self.__last_data_list) or not all([a is b for a, b in zip(self.__xdata_list, self.__last_data_list)]):
                Utility.fps_tick("update_"+self.__display_frame_rate_id)
                self.__last_data_list = copy.copy(self.__xdata_list)
        super().update()

    def update_regions(self, display, graphic_selection):
        dimensional_scales = display.displayed_dimensional_scales
        graphics = display.graphics

        self.__graphics = copy.copy(graphics)
        self.__graphic_selection = copy.copy(graphic_selection)

        if dimensional_scales is None or len(dimensional_scales) == 0:
            return

        data_scale = dimensional_scales[-1]
        dimensional_calibration = display.displayed_dimensional_calibrations[-1] if len(display.displayed_dimensional_calibrations) > 0 else Calibration.Calibration(scale=display.displayed_dimensional_scales[-1])

        def convert_to_calibrated_value_str(f):
            return u"{0}".format(dimensional_calibration.convert_to_calibrated_value_str(f, value_range=(0, data_scale), samples=data_scale, include_units=False))

        def convert_to_calibrated_size_str(f):
            return u"{0}".format(dimensional_calibration.convert_to_calibrated_size_str(f, value_range=(0, data_scale), samples=data_scale, include_units=False))

        regions = list()
        for graphic_index, graphic in enumerate(graphics):
            if isinstance(graphic, Graphics.IntervalGraphic):
                graphic_start, graphic_end = graphic.start, graphic.end
                graphic_start, graphic_end = min(graphic_start, graphic_end), max(graphic_start, graphic_end)
                left_channel = graphic_start * data_scale
                right_channel = graphic_end * data_scale
                left_text = convert_to_calibrated_value_str(left_channel)
                right_text = convert_to_calibrated_value_str(right_channel)
                middle_text = convert_to_calibrated_size_str(right_channel - left_channel)
                RegionInfo = collections.namedtuple("RegionInfo", ["channels", "selected", "index", "left_text", "right_text", "middle_text", "label", "style"])
                region = RegionInfo((graphic_start, graphic_end), graphic_selection.contains(graphic_index), graphic_index, left_text, right_text, middle_text, graphic.label, None)
                regions.append(region)
            elif isinstance(graphic, Graphics.ChannelGraphic):
                graphic_start, graphic_end = graphic.position, graphic.position
                graphic_start, graphic_end = min(graphic_start, graphic_end), max(graphic_start, graphic_end)
                left_channel = graphic_start * data_scale
                right_channel = graphic_end * data_scale
                left_text = convert_to_calibrated_value_str(left_channel)
                right_text = convert_to_calibrated_value_str(right_channel)
                middle_text = convert_to_calibrated_size_str(right_channel - left_channel)
                RegionInfo = collections.namedtuple("RegionInfo", ["channels", "selected", "index", "left_text", "right_text", "middle_text", "label", "style"])
                region = RegionInfo((graphic_start, graphic_end), graphic_selection.contains(graphic_index), graphic_index, left_text, right_text, middle_text, graphic.label, "tag")
                regions.append(region)

        self.__line_graph_regions_canvas_item.set_regions(regions)

    def handle_auto_display(self, display) -> bool:
        # enter key has been pressed
        data_and_metadata = display.data_and_metadata_for_display_panel
        display.view_to_selected_graphics(data_and_metadata)
        return True

    def prepare_display(self):
        """Prepare the display.

        This method gets called by the canvas layout/draw engine after being triggered by a call to `update`.

        When data or display parameters change, the internal state of the line plot gets updated. This method takes
        that internal state and updates the child canvas items.

        This method is always run on a thread and should be fast but doesn't need to be instant.
        """
        displayed_dimensional_calibration = self.__displayed_dimensional_calibration
        intensity_calibration = self.__intensity_calibration
        dimensional_calibration_style = self.__dimensional_calibration_style
        y_min = self.__y_min
        y_max = self.__y_max
        y_style = self.__y_style
        left_channel = self.__left_channel
        right_channel = self.__right_channel
        legend_labels = self.__legend_labels

        scalar_xdata_list = None

        def calculate_scalar_xdata(xdata_list):
            scalar_xdata_list = list()
            for xdata in xdata_list:
                if xdata:
                    scalar_data = Image.scalar_from_array(xdata.data)
                    scalar_data = Image.convert_to_grayscale(scalar_data)
                    scalar_intensity_calibration = dimensional_calibration_style.get_intensity_calibration(xdata)
                    scalar_dimensional_calibrations = dimensional_calibration_style.get_dimensional_calibrations(xdata.dimensional_shape, xdata.dimensional_calibrations)
                    # check whether there is common units between the data and the display
                    if not displayed_dimensional_calibration.units and self.__dimensional_calibration and scalar_dimensional_calibrations[-1].units == self.__dimensional_calibration.units:
                        scalar_dimensional_calibrations = scalar_dimensional_calibrations[0:-1] + [Calibration.Calibration(scale=self.__data_scale)]
                    if displayed_dimensional_calibration.units == scalar_dimensional_calibrations[-1].units:
                        scalar_xdata_list.append(DataAndMetadata.new_data_and_metadata(scalar_data, scalar_intensity_calibration, scalar_dimensional_calibrations))
                else:
                    scalar_xdata_list.append(None)
            return scalar_xdata_list

        data_scale = self.__data_scale
        xdata_list = self.__xdata_list

        if data_scale is not None:
            # update the line graph data
            left_channel = left_channel if left_channel is not None else 0
            right_channel = right_channel if right_channel is not None else data_scale
            left_channel, right_channel = min(left_channel, right_channel), max(left_channel, right_channel)

            scalar_data_list = None
            if y_min is None or y_max is None and len(xdata_list) > 0:
                scalar_xdata_list = calculate_scalar_xdata(xdata_list)
                scalar_data_list = [xdata.data if xdata else None for xdata in scalar_xdata_list]
            calibrated_data_min, calibrated_data_max, y_ticker = LineGraphCanvasItem.calculate_y_axis(scalar_data_list, y_min, y_max, intensity_calibration, y_style)
            axes = LineGraphCanvasItem.LineGraphAxes(data_scale, calibrated_data_min, calibrated_data_max, left_channel, right_channel, displayed_dimensional_calibration, intensity_calibration, y_style, y_ticker)

            if scalar_xdata_list is None:
                if len(xdata_list) > 0:
                    scalar_xdata_list = calculate_scalar_xdata(xdata_list)
                else:
                    scalar_xdata_list = list()

            line_graph_xdata_list = list()

            if self.__display_frame_rate_id:
                Utility.fps_tick("prepare_"+self.__display_frame_rate_id)

            for scalar_xdata in scalar_xdata_list:
                if scalar_xdata and scalar_xdata.is_data_1d:
                    line_graph_xdata_list.append(scalar_xdata)
                elif scalar_xdata and scalar_xdata.is_data_2d:
                    rows = min(16, scalar_xdata.dimensional_shape[0])
                    intensity_calibration = scalar_xdata.intensity_calibration
                    displayed_dimensional_calibration = scalar_xdata.dimensional_calibrations[-1]
                    for row in range(rows):
                        scalar_data_row = scalar_xdata.data[row:row + 1, :].reshape((scalar_xdata.dimensional_shape[-1],))
                        scalar_xdata_row = DataAndMetadata.new_data_and_metadata(scalar_data_row, intensity_calibration, [displayed_dimensional_calibration])
                        line_graph_xdata_list.append(scalar_xdata_row)
                else:
                    line_graph_xdata_list.append(None)

            line_graph_xdata_list = line_graph_xdata_list[0:16]

            colors = ('#1E90FF', "#F00", "#0F0", "#00F", "#FF0", "#0FF", "#F0F", "#888", "#800", "#080", "#008", "#CCC", "#880", "#088", "#808", "#964B00")

            for index, line_graph_canvas_item in enumerate(self.__line_graph_stack.canvas_items):
                if index < len(line_graph_xdata_list):
                    line_graph_canvas_item.set_is_filled(index == 0)
                    line_graph_canvas_item.set_color(colors[index])
                    line_graph_canvas_item.set_axes(axes)
                    line_graph_canvas_item.set_uncalibrated_xdata(line_graph_xdata_list[index])
                else:
                    line_graph_canvas_item.set_axes(None)
                    line_graph_canvas_item.set_uncalibrated_xdata(None)

            self.__line_graph_xdata_list = line_graph_xdata_list

            self.__update_canvas_items(axes, legend_labels)
        else:
            for line_graph_canvas_item in self.__line_graph_stack.canvas_items:
                line_graph_canvas_item.set_axes(None)
                line_graph_canvas_item.set_uncalibrated_xdata(None)
            self.__line_graph_xdata_list = list()
            self.__update_canvas_items(LineGraphCanvasItem.LineGraphAxes(), None)

    def _inserted(self, container):
        # make sure we get 'prepare_render' calls
        self.register_prepare_canvas_item(self)

    def _removed(self, container):
        # turn off 'prepare_render' calls
        self.unregister_prepare_canvas_item(self)

    def prepare_render(self):
        self.prepare_display()

    def _repaint(self, drawing_context):
        super()._repaint(drawing_context)

        if self.__display_frame_rate_id:
            Utility.fps_tick("display_"+self.__display_frame_rate_id)

        if self.__display_frame_rate_id:
            fps = Utility.fps_get("display_"+self.__display_frame_rate_id)
            fps2 = Utility.fps_get("frame_"+self.__display_frame_rate_id)
            fps3 = Utility.fps_get("update_"+self.__display_frame_rate_id)
            fps4 = Utility.fps_get("prepare_"+self.__display_frame_rate_id)

            rect = self.canvas_bounds

            with drawing_context.saver():
                font = "normal 11px serif"
                text_pos = Geometry.IntPoint(y=rect[0][0], x=rect[0][1] + rect[1][1] - 100)
                drawing_context.begin_path()
                drawing_context.move_to(text_pos.x, text_pos.y)
                drawing_context.line_to(text_pos.x + 120, text_pos.y)
                drawing_context.line_to(text_pos.x + 120, text_pos.y + 60)
                drawing_context.line_to(text_pos.x, text_pos.y + 60)
                drawing_context.close_path()
                drawing_context.fill_style = "rgba(255, 255, 255, 0.6)"
                drawing_context.fill()
                drawing_context.font = font
                drawing_context.text_baseline = "middle"
                drawing_context.text_align = "left"
                drawing_context.fill_style = "#000"
                drawing_context.fill_text("display:" + fps, text_pos.x + 8, text_pos.y + 10)
                drawing_context.fill_text("frame:" + fps2, text_pos.x + 8, text_pos.y + 30)
                drawing_context.fill_text("update:" + fps3, text_pos.x + 8, text_pos.y + 50)
                drawing_context.fill_text("prepare:" + fps4, text_pos.x + 8, text_pos.y + 70)

    def __update_canvas_items(self, axes, legend_labels):
        self.__line_graph_background_canvas_item.set_axes(axes)
        self.__line_graph_regions_canvas_item.set_axes(axes)
        self.__line_graph_regions_canvas_item.set_calibrated_data(self.line_graph_canvas_item.calibrated_xdata.data if self.line_graph_canvas_item and self.line_graph_canvas_item.calibrated_xdata else None)
        self.__line_graph_frame_canvas_item.set_draw_frame(axes.is_valid)
        self.__line_graph_legend_canvas_item.set_legend_labels(legend_labels)
        self.__line_graph_vertical_axis_label_canvas_item.set_axes(axes)
        self.__line_graph_vertical_axis_scale_canvas_item.set_axes(axes, self.__get_font_metrics_fn)
        self.__line_graph_vertical_axis_ticks_canvas_item.set_axes(axes)
        self.__line_graph_horizontal_axis_label_canvas_item.set_axes(axes)
        self.__line_graph_horizontal_axis_scale_canvas_item.set_axes(axes)
        self.__line_graph_horizontal_axis_ticks_canvas_item.set_axes(axes)

        self.__axes = axes

    def mouse_entered(self):
        if super().mouse_entered():
            return True
        self.__mouse_in = True
        return True

    def mouse_exited(self):
        if super().mouse_exited():
            return True
        self.__mouse_in = False
        self.__update_cursor_info()
        if self.delegate:  # allow display to work without delegate
            # whenever the cursor exits, clear the cursor display
            self.delegate.cursor_changed(None)
        return True

    def mouse_double_clicked(self, x, y, modifiers):
        if super().mouse_clicked(x, y, modifiers):
            return True
        if self.delegate.tool_mode == "pointer":
            pos = Geometry.IntPoint(x=x, y=y)
            if self.line_graph_horizontal_axis_group_canvas_item.canvas_bounds.contains_point(self.map_to_canvas_item(pos, self.line_graph_horizontal_axis_group_canvas_item)):
                command = self.delegate.create_change_display_command()
                self.delegate.update_display_properties({"left_channel": None, "right_channel": None})
                self.delegate.push_undo_command(command)
                return True
            elif self.__line_graph_vertical_axis_group_canvas_item.canvas_bounds.contains_point(self.map_to_canvas_item(pos, self.__line_graph_vertical_axis_group_canvas_item)):
                command = self.delegate.create_change_display_command()
                self.delegate.update_display_properties({"y_min": None, "y_max": None})
                self.delegate.push_undo_command(command)
                return True
        return False

    def mouse_position_changed(self, x, y, modifiers):
        if super().mouse_position_changed(x, y, modifiers):
            return True
        if self.delegate.tool_mode == "pointer":
            if self.__axes and self.__axes.is_valid:
                pos = Geometry.IntPoint(x=x, y=y)
                if self.line_graph_horizontal_axis_group_canvas_item.canvas_bounds.contains_point(self.map_to_canvas_item(pos, self.line_graph_horizontal_axis_group_canvas_item)):
                    if modifiers.control:
                        self.cursor_shape = "split_horizontal"
                    else:
                        self.cursor_shape = "hand"
                elif self.__line_graph_vertical_axis_group_canvas_item.canvas_bounds.contains_point(self.map_to_canvas_item(pos, self.__line_graph_vertical_axis_group_canvas_item)):
                    if modifiers.control:
                        self.cursor_shape = "split_vertical"
                    else:
                        self.cursor_shape = "hand"
                else:
                    self.cursor_shape = "arrow"
        elif self.delegate.tool_mode == "interval":
            self.cursor_shape = "cross"
        self.__last_mouse = Geometry.IntPoint(x=x, y=y)
        self.__update_cursor_info()
        new_rescale = modifiers.control
        if self.__tracking_horizontal and self.__tracking_rescale != new_rescale:
            self.end_tracking(modifiers)
            self.begin_tracking_horizontal(Geometry.IntPoint(x=x, y=y), rescale=new_rescale)
        elif self.__tracking_vertical and self.__tracking_rescale != new_rescale:
            self.end_tracking(modifiers)
            self.begin_tracking_vertical(Geometry.IntPoint(x=x, y=y), rescale=new_rescale)
        return self.continue_tracking(Geometry.IntPoint(x=x, y=y), modifiers)

    def mouse_pressed(self, x, y, modifiers):
        if super().mouse_pressed(x, y, modifiers):
            return True
        if not self.__axes or not self.__axes.is_valid:
            return False
        self.__undo_command = None
        pos = Geometry.IntPoint(x=x, y=y)
        self.delegate.begin_mouse_tracking()
        if self.delegate.tool_mode == "pointer":
            if self.__line_graph_regions_canvas_item.canvas_bounds.contains_point(self.map_to_canvas_item(pos, self.__line_graph_regions_canvas_item)):
                self.begin_tracking_regions(pos, modifiers)
                return True
            elif self.line_graph_horizontal_axis_group_canvas_item.canvas_bounds.contains_point(self.map_to_canvas_item(pos, self.line_graph_horizontal_axis_group_canvas_item)):
                self.begin_tracking_horizontal(pos, rescale=modifiers.control)
                return True
            elif self.__line_graph_vertical_axis_group_canvas_item.canvas_bounds.contains_point(self.map_to_canvas_item(pos, self.__line_graph_vertical_axis_group_canvas_item)):
                self.begin_tracking_vertical(pos, rescale=modifiers.control)
                return True
        if self.delegate.tool_mode == "interval":
            if self.__line_graph_regions_canvas_item.canvas_bounds.contains_point(self.map_to_canvas_item(pos, self.__line_graph_regions_canvas_item)):
                widget_mapping = self.__get_mouse_mapping()
                x = widget_mapping.map_point_widget_to_channel_norm(pos)
                region = Graphics.IntervalGraphic()
                region.start = x
                region.end = x
                self.__undo_command = self.delegate.add_and_select_region(region)
                self.begin_tracking_regions(pos, Graphics.NullModifiers())
                return True
        return False

    def mouse_released(self, x, y, modifiers):
        if super().mouse_released(x, y, modifiers):
            return True
        if not self.__axes or not self.__axes.is_valid:
            return False
        self.end_tracking(modifiers)
        self.delegate.end_mouse_tracking(self.__undo_command)
        return False

    def _mouse_dragged(self, start: float, end: float, modifiers=None) -> None:
        # for testing
        plot_origin = self.line_graph_canvas_item.map_to_canvas_item(Geometry.IntPoint(), self)
        plot_left = self.line_graph_canvas_item.canvas_rect.left + plot_origin.x
        plot_width = self.line_graph_canvas_item.canvas_rect.width
        modifiers = modifiers if modifiers else CanvasItem.KeyboardModifiers()
        self.mouse_pressed(plot_left + plot_width * start, 100, modifiers)
        self.mouse_position_changed(plot_left + plot_width * start, 100, modifiers)
        self.mouse_position_changed(plot_left + plot_width * (start + end) / 2, 100, modifiers)
        self.mouse_position_changed(plot_left + plot_width * end, 100, modifiers)
        self.mouse_released(plot_left + plot_width * end, 100, modifiers)

    def context_menu_event(self, x, y, gx, gy):
        return self.delegate.show_display_context_menu(gx, gy)

    # ths message comes from the widget
    def key_pressed(self, key):
        if super().key_pressed(key):
            return True
        # only handle keys if we're directly embedded in an image panel
        if key.is_delete:
            self.delegate.delete_key_pressed()
            return True
        if key.is_enter_or_return:
            self.delegate.enter_key_pressed()
            return True
        if key.is_arrow:
            mapping = self.__get_mouse_mapping()
            amount = 10.0 if key.modifiers.shift else 1.0
            if key.is_left_arrow:
                self.delegate.nudge_selected_graphics(mapping, Geometry.FloatPoint(y=0, x=-amount))
            elif key.is_right_arrow:
                self.delegate.nudge_selected_graphics(mapping, Geometry.FloatPoint(y=0, x=amount))
            return True
        if key.key == 70 and key.modifiers.shift and key.modifiers.alt:
            if self.__display_frame_rate_id is None:
                self.__display_frame_rate_id = str(id(self))
            else:
                self.__display_frame_rate_id = None
            return True
        return False

    def __get_mouse_mapping(self):
        data_scale = self.__data_scale
        plot_origin = self.__line_graph_regions_canvas_item.map_to_canvas_item(Geometry.IntPoint(), self)
        plot_rect = self.__line_graph_regions_canvas_item.canvas_bounds.translated(plot_origin)
        left_channel = self.__axes.drawn_left_channel
        right_channel = self.__axes.drawn_right_channel
        return LinePlotCanvasItemMapping(data_scale, plot_rect, left_channel, right_channel)

    def begin_tracking_regions(self, pos, modifiers):
        # keep track of general drag information
        self.__graphic_drag_start_pos = Geometry.IntPoint.make(pos)
        self.__graphic_drag_changed = False
        if self.__axes and self.__axes.is_valid:
            self.__tracking_selections = True
            graphics = self.__graphics
            selection_indexes = self.__graphic_selection.indexes
            for graphic_index, graphic in enumerate(graphics):
                if isinstance(graphic, (Graphics.IntervalGraphic, Graphics.ChannelGraphic)):
                    already_selected = graphic_index in selection_indexes
                    multiple_items_selected = len(selection_indexes) > 1
                    move_only = not already_selected or multiple_items_selected
                    widget_mapping = self.__get_mouse_mapping()
                    part, specific = graphic.test(widget_mapping, self.__get_font_metrics_fn, self.__graphic_drag_start_pos, move_only)
                    if part:
                        # select item and prepare for drag
                        self.graphic_drag_item_was_selected = already_selected
                        if not self.graphic_drag_item_was_selected:
                            if modifiers.control:
                                self.delegate.add_index_to_selection(graphic_index)
                                selection_indexes.add(graphic_index)
                            else:
                                self.delegate.set_selection(graphic_index)
                                selection_indexes.clear()
                                selection_indexes.add(graphic_index)
                        # keep track of info for the specific item that was clicked
                        self.__graphic_drag_item = graphic
                        self.__graphic_drag_part = part
                        # keep track of drag information for each item in the set
                        self.__graphic_drag_indexes = selection_indexes
                        for index in self.__graphic_drag_indexes:
                            graphic = graphics[index]
                            self.__graphic_drag_items.append(graphic)
                            self.__graphic_part_data[index] = graphic.begin_drag()
                        break

    def begin_tracking_horizontal(self, pos, rescale):
        plot_origin = self.line_graph_horizontal_axis_group_canvas_item.map_to_canvas_item(Geometry.IntPoint(), self)
        plot_rect = self.line_graph_horizontal_axis_group_canvas_item.canvas_bounds.translated(plot_origin)
        self.__tracking_horizontal = True
        self.__tracking_rescale = rescale
        self.__tracking_start_pos = pos
        self.__tracking_start_left_channel = self.__axes.drawn_left_channel
        self.__tracking_start_right_channel = self.__axes.drawn_right_channel
        self.__tracking_start_drawn_channel_per_pixel = float(self.__tracking_start_right_channel - self.__tracking_start_left_channel) / plot_rect.width
        self.__tracking_start_origin_pixel = self.__tracking_start_pos.x - plot_rect.left
        self.__tracking_start_channel = self.__tracking_start_left_channel + self.__tracking_start_origin_pixel * self.__tracking_start_drawn_channel_per_pixel

    def begin_tracking_vertical(self, pos, rescale):
        plot_height = self.__line_graph_area_stack.canvas_bounds.height - 1
        self.__tracking_vertical = True
        self.__tracking_rescale = rescale
        self.__tracking_start_pos = pos
        self.__tracking_start_calibrated_data_min = self.__axes.calibrated_data_min
        self.__tracking_start_calibrated_data_max = self.__axes.calibrated_data_max
        self.__tracking_start_calibrated_data_per_pixel = (self.__tracking_start_calibrated_data_max - self.__tracking_start_calibrated_data_min) / plot_height
        plot_origin = self.__line_graph_vertical_axis_group_canvas_item.map_to_canvas_item(Geometry.IntPoint(), self)
        plot_rect = self.__line_graph_vertical_axis_group_canvas_item.canvas_bounds.translated(plot_origin)
        if 0.0 >= self.__tracking_start_calibrated_data_min and 0.0 <= self.__tracking_start_calibrated_data_max:
            calibrated_unit_per_pixel = (self.__tracking_start_calibrated_data_max - self.__tracking_start_calibrated_data_min) / (plot_rect.height - 1)
            origin_offset_pixels = (0.0 - self.__tracking_start_calibrated_data_min) / calibrated_unit_per_pixel
            calibrated_origin = self.__tracking_start_calibrated_data_min + origin_offset_pixels * self.__tracking_start_calibrated_data_per_pixel
            self.__tracking_start_origin_y = origin_offset_pixels
            self.__tracking_start_calibrated_origin = calibrated_origin
        else:
            self.__tracking_start_origin_y = 0  # the distance the origin is up from the bottom
            self.__tracking_start_calibrated_origin = self.__tracking_start_calibrated_data_min

    def continue_tracking(self, pos, modifiers):
        if self.__tracking_selections:
            if self.__graphic_drag_item is None and not self.__graphic_drag_changed:
                pos = self.__graphic_drag_start_pos
                widget_mapping = self.__get_mouse_mapping()
                x = widget_mapping.map_point_widget_to_channel_norm(pos)
                region = Graphics.IntervalGraphic()
                region.start = x
                region.end = x
                self.__undo_command = self.delegate.add_and_select_region(region)
                selection_indexes = self.__graphic_selection.indexes
                for graphic_index, graphic in enumerate(self.__graphics):
                    if graphic == region:
                        part, specific = graphic.test(widget_mapping, self.__get_font_metrics_fn, self.__graphic_drag_start_pos, False)
                        if part:
                            self.graphic_drag_item_was_selected = False
                            self.delegate.set_selection(graphic_index)
                            selection_indexes.clear()
                            selection_indexes.add(graphic_index)
                            self.__graphic_drag_item = graphic
                            self.__graphic_drag_part = part
                            self.__graphic_drag_indexes = selection_indexes
                            self.__graphic_drag_items.append(graphic)
                            self.__graphic_part_data[graphic_index] = graphic.begin_drag()
            else:
                if not self.__undo_command:
                    self.__undo_command = self.delegate.create_change_graphics_command()
            # x,y already have transform applied
            self.__last_mouse = copy.copy(pos)
            self.__update_cursor_info()
            if self.__graphic_drag_items:
                widget_mapping = self.__get_mouse_mapping()
                self.delegate.update_graphics(widget_mapping, self.__graphic_drag_items, self.__graphic_drag_part,
                                              self.__graphic_part_data, self.__graphic_drag_start_pos, pos, modifiers)
                self.__graphic_drag_changed = True
                self.__line_graph_regions_canvas_item.update()
        elif self.__tracking_horizontal:
            if not self.__undo_command:
                self.__undo_command = self.delegate.create_change_display_command()
            if self.__tracking_rescale:
                plot_origin = self.line_graph_horizontal_axis_group_canvas_item.map_to_canvas_item(Geometry.IntPoint(), self)
                plot_rect = self.line_graph_horizontal_axis_group_canvas_item.canvas_bounds.translated(plot_origin)
                pixel_offset_x = pos.x - self.__tracking_start_pos.x
                scaling = math.pow(10, pixel_offset_x/96.0)  # 10x per inch of travel, assume 96dpi
                new_drawn_channel_per_pixel = self.__tracking_start_drawn_channel_per_pixel / scaling
                left_channel = int(round(self.__tracking_start_channel - new_drawn_channel_per_pixel * self.__tracking_start_origin_pixel))
                right_channel = int(round(self.__tracking_start_channel + new_drawn_channel_per_pixel * (plot_rect.width - self.__tracking_start_origin_pixel)))
                self.delegate.update_display_properties({"left_channel": left_channel, "right_channel": right_channel})
                return True
            else:
                delta = pos - self.__tracking_start_pos
                left_channel = int(self.__tracking_start_left_channel - self.__tracking_start_drawn_channel_per_pixel * delta.x)
                right_channel = int(self.__tracking_start_right_channel - self.__tracking_start_drawn_channel_per_pixel * delta.x)
                self.delegate.update_display_properties({"left_channel": left_channel, "right_channel": right_channel})
                return True
        elif self.__tracking_vertical:
            if not self.__undo_command:
                self.__undo_command = self.delegate.create_change_display_command()
            if self.__tracking_rescale:
                plot_origin = self.__line_graph_vertical_axis_group_canvas_item.map_to_canvas_item(Geometry.IntPoint(), self)
                plot_rect = self.__line_graph_vertical_axis_group_canvas_item.canvas_bounds.translated(plot_origin)
                origin_y = plot_rect.bottom - 1 - self.__tracking_start_origin_y  # pixel position of y-origin
                calibrated_offset = self.__tracking_start_calibrated_data_per_pixel * (origin_y - self.__tracking_start_pos.y)
                pixel_offset = origin_y - pos.y
                pixel_offset = max(pixel_offset, 1) if origin_y > self.__tracking_start_pos.y else min(pixel_offset, -1)
                new_calibrated_data_per_pixel = calibrated_offset / pixel_offset
                calibrated_data_min = self.__tracking_start_calibrated_origin - new_calibrated_data_per_pixel * self.__tracking_start_origin_y
                calibrated_data_max = self.__tracking_start_calibrated_origin + new_calibrated_data_per_pixel * (plot_rect.height - 1 - self.__tracking_start_origin_y)
                uncalibrated_data_min = self.__axes.uncalibrate_y(calibrated_data_min)
                uncalibrated_data_max = self.__axes.uncalibrate_y(calibrated_data_max)
                self.delegate.update_display_properties({"y_min": uncalibrated_data_min, "y_max": uncalibrated_data_max})
                return True
            else:
                delta = pos - self.__tracking_start_pos
                calibrated_data_min = self.__tracking_start_calibrated_data_min + self.__tracking_start_calibrated_data_per_pixel * delta.y
                calibrated_data_max = self.__tracking_start_calibrated_data_max + self.__tracking_start_calibrated_data_per_pixel * delta.y
                uncalibrated_data_min = self.__axes.uncalibrate_y(calibrated_data_min)
                uncalibrated_data_max = self.__axes.uncalibrate_y(calibrated_data_max)
                self.delegate.update_display_properties({"y_min": uncalibrated_data_min, "y_max": uncalibrated_data_max})
                return True
        return False

    def end_tracking(self, modifiers):
        if not self.__graphic_drag_items and not modifiers.control:
            self.delegate.clear_selection()
        if self.__tracking_selections:
            if self.__axes and self.__axes.is_valid:
                graphics = self.__graphics
                for index in self.__graphic_drag_indexes:
                    graphic = graphics[index]
                    graphic.end_drag(self.__graphic_part_data[index])
                if self.__graphic_drag_items and not self.__graphic_drag_changed:
                    graphic_index = graphics.index(self.__graphic_drag_item)
                    # user didn't move graphic
                    if not modifiers.control:
                        # user clicked on a single graphic
                        self.delegate.set_selection(graphic_index)
                    else:
                        # user control clicked. toggle selection
                        # if control is down and item is already selected, toggle selection of item
                        if self.graphic_drag_item_was_selected:
                            self.delegate.remove_index_from_selection(graphic_index)
                        else:
                            self.delegate.add_index_to_selection(graphic_index)
            self.__graphic_drag_items = []
            self.__graphic_drag_item = None
            self.__graphic_part_data = {}
            self.__graphic_drag_indexes = []
        self.delegate.tool_mode = "pointer"
        self.__tracking_horizontal = False
        self.__tracking_vertical = False
        self.__tracking_selections = False
        self.prepare_display()

    def __update_cursor_info(self):
        """ Map the mouse to the 1-d position within the line graph. """

        if not self.delegate:  # allow display to work without delegate
            return

        if self.__mouse_in and self.__last_mouse:
            pos_1d = None
            axes = self.__axes
            line_graph_canvas_item = self.line_graph_canvas_item
            if axes and axes.is_valid and line_graph_canvas_item:
                mouse = self.map_to_canvas_item(self.__last_mouse, line_graph_canvas_item)
                plot_rect = line_graph_canvas_item.canvas_bounds
                if plot_rect.contains_point(mouse):
                    mouse = mouse - plot_rect.origin
                    x = float(mouse.x) / plot_rect.width
                    px = axes.drawn_left_channel + x * (axes.drawn_right_channel - axes.drawn_left_channel)
                    pos_1d = px,
            self.delegate.cursor_changed(pos_1d)
